from openagi.planner.task_decomposer import TaskPlanner
from openagi.actions.tools.ddg_search import DuckDuckGoSearch
from openagi.actions.tools.tavilyqasearch import TavilyWebSearchQA
from openagi.agent import Admin
from openagi.actions.tools.serper_search import SerperSearch
from openagi.actions.tools.exasearch import ExaSearch
from openagi.actions.tools.serp_search import GoogleSerpAPISearch
from openagi.llms.gemini import GeminiModel
from openagi.worker import Worker

import os
os.environ['GOOGLE_API_KEY'] = ""
os.environ['Gemini_MODEL'] = "gemini-1.5-flash"
os.environ['Gemini_TEMP'] = "0.7"

gemini_config = GeminiModel.load_from_env_config()
llm = GeminiModel(config=gemini_config)

plan = TaskPlanner(autonomous=False,human_intervene=False)

#SerperSearch.set_config({"api_key":""})
#TavilyWebSearchQA.set_config({"api_key": "tvly-"})
#ExaSearch.set_config({"api_key":"---"})
#GoogleSerpAPISearch.set_config({"api_key":""})

cricket_worker = Worker(
    role = "Cricket Analyst and Finder",
    instructions = "You need to analyse the user query and gather the information on what is required. You only trust on the source you get",
    actions = [TavilyWebSearchQA]    
)

admin = Admin(
    actions = [TavilyWebSearchQA],
    planner = plan,
    llm = llm,
)
admin.assign_workers([cricket_worker])

res = admin.run(
    query="how many runs did Rohit sharma scored in 3rd ODI match against Sri Lanka during Sri Lanka tour 2024",
    description=f"give me the results of Rohit sharma scores during India vs Sri Lanka ODI 2024 series",
)
print(res)


from typing import Dict, List, Optional
from pydantic import Field
from openagi.prompts.base import BasePrompt
from openagi.prompts.constants import FAILURE_VARS

start = FAILURE_VARS["start"]
end = FAILURE_VARS["end"]

task_execution = """
You are an expert Task executor and skillful problem solver \
You are loyal to your job and execute the task with 100 percent accuracy \ 
Your primary role is to clearly understand the Task Objective to provide optimal results using the supported actions. \ 
Below is a list of tasks that need to be executed:

You can code if needs be. 

All Tasks:
{all_tasks}

You are provided with the current task details from the user.
Current Task:
Name: {current_task_name}
Description: {current_description}

To execute the current task, refer to the details of the Previous Task and the All Tasks provided.

Previous Task:
{previous_task}

Supported Actions:
{supported_actions}

Your task is to understand and return a JSON array with the actions to be executed along with the values for each parameter. Use only the Supported Actions. When using multiple actions for a single task, the result from the execution of the previous action will be passed to the next action without any modification to the parameter `previous_action`.

Task Objective:
{objective}

The output should be a markdown code snippet formatted in the following schema, including the leading and trailing "```json" and "```":

```json
[
    {
        "cls": {"kls": "<action>", "module": "<module>"},
        "params": {
            "description": "<description>",
            "name": "<name>",
            "filename": "<filename>",
            "file_content": "<file_content>",
            "file_mode": "w"
        }
    }
]
```

If the task cannot be executed using the available actions, return the failure reason within the delimiters $start$ and $end$ as shown below and provide some guidance on what type of generic actions would help in acheiving it:
$start$ Couldn't execute the `{current_task_name}` task. $end$
"""

# In order to retreive them(previous task results of the current objective) just use ```MemoryRagAction```.


task_execution = task_execution.replace("$start$", start)
task_execution = task_execution.replace("$end$", end)


class TaskExecutor(BasePrompt):
    objective: str = Field(..., description="Final objective")
    all_tasks: List[Dict] = Field(
        ..., description="List of tasks to be executed that was generated earlier"
    )
    current_task_name: str = Field(..., description="Current task name to be executed.")
    current_description: str = Field(..., description="Current task name to be executed.")
    previous_task: Optional[str] = Field(..., description="Previous task, description & result.")
    supported_actions: List[Dict] = Field(
        ...,
        description="Supported Actions that can be used to acheive the current task.",
    )
    base_prompt: str = task_execution



from typing import Any
from pydantic import Field
from openagi.actions.base import BaseAction
from openagi.prompts.summarizer import SummarizerPrompt


class SummarizerAction(BaseAction):
    """Summarizer Action"""

    past_messages: Any = Field(
        ...,
        description="Messages/Data to be summarized",
    )

    def execute(self):
        summarizer: str = SummarizerPrompt.from_template({"past_messages": self.past_messages})
        return self.llm.run(summarizer)



from typing import Dict
from pydantic import BaseModel, Field


class BasePrompt(BaseModel):
    name: str = Field(default="BasePrompt", description="Name of the prompt.")
    description: str = Field(
        default="BasePrompt class to be used by other actions that get created.",
        description="Description of the prompt.",
    )
    base_prompt: str = Field(default_factory=str, description="Base prompt to be used.")

    def get_prompt(self):
        raise NotImplementedError("Subclasses must implement this method.")

    @classmethod
    def prompt_variables(cls):
        return {
            field_name: field.field_info.description
            for field_name, field in cls.model_fields.items()
        }

    @classmethod
    def from_template(cls, variables: Dict):
        x = cls(**variables)
        for k, v in variables.items():
            placeholder = "{" + f"{k}" + "}"
            x.base_prompt = x.base_prompt.replace(placeholder, f"{v}")
        return x.base_prompt



import json
from typing import Any,Dict
from openagi.actions.base import BaseAction
from pydantic import Field
from duckduckgo_search import DDGS
import logging

class DuckDuckGoSearch(BaseAction):
    """Use this Action to search DuckDuckGo for a query."""

    name: str = Field(
        default_factory=str,
        description="DuckDuckGoSearch Action to search over duckduckgo using the query.",
    )
    description: str = Field(
        default_factory=str,
        description="This action is used to search for words, documents, images, videos, news, maps and text translation using the DuckDuckGo.com search engine.",
    )

    query: Any = Field(
        default_factory=str,
        description="User query, a string, to fetch web search results from DuckDuckGo",
    )

    max_results: int = Field(
        default=10,
        description="Total results, in int, to be executed from the search. Defaults to 10.",
    )

    def _get_ddgs(self):
        return DDGS()

    def execute(self):
        if self.max_results > 15:
            logging.info("Over threshold value... Limiting the Max results to 15")
            self.max_results = 15
        
        result = self._get_ddgs().text(
            self.query,
            max_results=self.max_results,
        )
        return json.dumps(result)
    
    @classmethod
    def from_user_config(cls, config: Dict[str, Any]) -> 'DuckDuckGoSearch':
        """
        Create a DuckDuckGoSearch instance from a user-provided configuration dictionary.
        
        :param config: A dictionary containing user-specified configuration.
        :return: An instance of DuckDuckGoSearch.
        """
        return cls(**config)


class DuckDuckGoNewsSearch(DuckDuckGoSearch):
    """Use this Action to get the latest news from DuckDuckGo."""

    def execute(self):
        ddgs = self._get_ddgs()
        return json.dumps(
            ddgs.news(keywords=self.query, max_results=(self.max_results)), indent=2
        )

if __name__ == "__main__":
    user_config = {
        "query": "Python programming",
        "max_results": 5
    }
    action = DuckDuckGoSearch
    print(action.execute("query",max_results=15))


from openagi.actions.base import BaseAction
from pydantic import Field
from openagi.exception import OpenAGIException
from typing import ClassVar, Dict, Any

try:
    from tavily import TavilyClient
except ImportError:
    raise OpenAGIException("Install Tavily with cmd `pip install tavily-python`")
    
class ConfigurableAction(BaseAction):
    config: ClassVar[Dict[str, Any]] = {}

    @classmethod
    def set_config(cls, *args, **kwargs):
        if args:
            if len(args) == 1 and isinstance(args[0], dict):
                cls.config.update(args[0])
            else:
                raise ValueError("If using positional arguments, a single dictionary must be provided.")
        cls.config.update(kwargs)

    @classmethod
    def get_config(cls, key: str, default: Any = None) -> Any:
        return cls.config.get(key, default)
    
class TavilyWebSearchQA(ConfigurableAction):
    """
    Tavily Web Search QA is a tool used when user needs to ask the question in terms of query to get response
    """
    query: str = Field(..., description="User query or question")

    def execute(self):
        api_key = self.get_config('api_key')
        if not api_key:
            raise OpenAGIException("API KEY NOT FOUND. Use TavilyWebSearchQA.set_config(api_key='your_key') to set the API key.")
        
        client = TavilyClient(api_key=api_key)
        response = client.qna_search(query=self.query)
        return response


from queue import Queue
from typing import Dict, List, Optional
from openagi.tasks.task import Task


class TaskLists:
    def __init__(self) -> None:
        self.tasks = Queue()
        self.completed_tasks = Queue()

    def add_task(self, task: Task) -> None:
        """Adds a Task instance to the queue."""
        self.tasks.put(task)

    def add_tasks(self, tasks: List[Dict[str, str]]):
        for task in tasks:
            task["name"] = task["task_name"]
            worker_config: Optional[Dict[str, str]] = None
            
            if all(key in task for key in ["role", "instruction", "worker_name", "supported_actions"]):
                worker_config = {
                    "role": task["role"],
                    "instructions": task["instruction"],
                    "name": task["worker_name"],
                    "supported_actions": task["supported_actions"]
                }
            task["worker_config"] = worker_config
            self.add_task(Task(**task))

    def get_tasks_queue(self) -> List:
        return self.tasks

    def get_tasks_lists(self):
        return [dict(task.model_fields.items()) for task in list(self.tasks.queue)]

    def get_next_unprocessed_task(self) -> Task:
        """Retrieves the next unprocessed task from the queue."""
        if not self.tasks.empty():
            return self.tasks.get_nowait()
        return None

    @property
    def all_tasks_completed(self) -> bool:
        """Checks if all tasks in the queue have been processed."""
        return self.tasks.empty()

    def add_completed_tasks(self, task: Task):
        self.completed_tasks.put(task)



from typing import Any
from langchain_core.messages import HumanMessage
from openagi.exception import OpenAGIException
from openagi.llms.base import LLMBaseModel, LLMConfigModel
from openagi.utils.yamlParse import read_from_env

try:
   from langchain_groq import ChatGroq
except ImportError:
  raise OpenAGIException("Install langchain groq with cmd `pip install langchain-groq`")

class GroqConfigModel(LLMConfigModel):
    """Configuration model for Groq Chat model."""

    groq_api_key: str
    model_name: str = "mixtral-8x7b-32768"
    temperature: float = 0.1

class GroqModel(LLMBaseModel):
    """Chat Groq Model implementation of the LLMBaseModel.

    This class implements the specific logic required to work with Chat Groq Model.
    """

    config: Any

    def load(self):
        """Initializes the GroqModel instance with configurations."""
        self.llm = ChatGroq(
            model_name = self.config.model_name,
            groq_api_key = self.config.groq_api_key,
            temperature = self.config.temperature
        )
        return self.llm
    
    def run(self, input_data: str):
        """Runs the Chat Groq model with the provided input text.

        Args:
            input_data: The input text to process.

        Returns:
            The response from Groq model with low inference latency.
        """
        if not self.llm:
            self.load()
        if not self.llm:
            raise ValueError("`llm` attribute not set.")
        message = HumanMessage(content=input_data)
        resp = self.llm([message])
        return resp.content
    
    @staticmethod
    def load_from_env_config() -> GroqConfigModel:
        """Loads the GroqModel configurations from a env file.

        Returns:
            An instance of GroqConfigModel with loaded configurations.
        """
        return GroqConfigModel(
            groq_api_key=read_from_env("GROQ_API_KEY", raise_exception=True),
            model_name = read_from_env("GROQ_MODEL",raise_exception=True),
            temperature=read_from_env("GROQ_TEMP",raise_exception=True)
        )



import logging
import os
from typing import Any
from pydantic import Field, field_validator
from serpapi import GoogleSearch
from typing import ClassVar, Dict, Any
from openagi.actions.base import BaseAction
from openagi.exception import OpenAGIException

class ConfigurableAction(BaseAction):
    config: ClassVar[Dict[str, Any]] = {}

    @classmethod
    def set_config(cls, *args, **kwargs):
        if args:
            if len(args) == 1 and isinstance(args[0], dict):
                cls.config.update(args[0])
            else:
                raise ValueError("If using positional arguments, a single dictionary must be provided.")
        cls.config.update(kwargs)

    @classmethod
    def get_config(cls, key: str, default: Any = None) -> Any:
        return cls.config.get(key, default)
    
class GoogleSerpAPISearch(ConfigurableAction):
    """Google Serp API Search Tool"""

    query: str = Field(
        ..., description="User query of type string used to fetch web search results from Google."
    )
    max_results: Any = Field(
        default=10,
        description="Total results, an integer, to be executed from the search. Defaults to 10",
    )
    
    @field_validator("max_results")
    @classmethod
    def actions_validator(cls, max_results):
        if not max_results or not isinstance(max_results, int):
            logging.warning("Max Results set to 10(default).")
            max_results = 10
        return max_results

    def execute(self):
        api_key = self.get_config('api_key')
        if not api_key:
            raise OpenAGIException("API KEY NOT FOUND. Use GoogleSerpAPISearch.set_config(api_key='your_key') to set the API key.")

        search_dict = {
            "q": self.query,
            "hl": "en",
            "gl": "us",
            "num": self.max_results,
            "api_key": api_key,
        }
        logging.debug(f"{search_dict=}")
        search = GoogleSearch(search_dict)

        max_retries = 3
        retries = 1
        result = None

        while retries < max_retries and not result:
            try:
                result = search.get_dict()
            except TypeError:
                logging.error("Error during GoogleSearch.", exc_info=True)
                continue
            retries += 1

        if not result:
            raise OpenAGIException(f"Unable to generate result for the query {self.query}")

        logging.debug(result)
        logging.info(f"NOTE: REMOVE THIS BEFORE RELEASE:\n{result}\n")
        if error := result.get("error", NotImplemented):
            raise OpenAGIException(
                f"Error while running action {self.__class__.__name__}: {error}"
            )

        meta_data = ""
        for info in result.get("organic_results"):
            meta_data += f"CONTEXT: {info['title']} \ {info['snippet']}"
            meta_data += f"Reference URL: {info['link']}"
        return meta_data


from pydantic import Field

from openagi.actions.base import BaseAction


class HumanCLIInput(BaseAction):
    ques_prompt: str = Field(
        default="Do you think this task is progressing as expected [y/n] or [yes/no]: ",
        description="question to be asked to human",
    )

    def execute(self):
        response = input(f"Agent: {self.ques_prompt}\nYou: ")
        return response



from typing import Optional, Dict, Any

from pydantic import BaseModel, Field

from openagi.utils.helper import get_default_id


class Task(BaseModel):
    id: str = Field(default_factory=get_default_id)
    name: str = Field(..., description="Name of task being.")
    description: str = Field(..., description="Description of the individual task.")
    result: Optional[str] = Field(..., default_factory=str, description="Result of the task.")
    actions: Optional[str] = Field(
        ...,
        default_factory=str,
        description="Actions undertaken to acheieve the task. Usually set after the current task is executed.",
    )
    worker_id: Optional[str] = Field(
        description="WorkerId associated to accomplish the given task using supported actions.",
        default_factory=str,
    )
    worker_config: Optional[Dict[str, Any]] = Field(
        description="Stores workers configuration values"
    )
    @property
    def is_done(self):
        return bool(self.result)

    def set_result(self, result):
        self.result = result



from pprint import pprint

from pydantic import Field

from openagi.actions.base import BaseAction


class ConsolePrint(BaseAction):
    content: str = Field(
        ...,
        description="The content/data passed will be logged into the console using pprint.pprint() module.",
    )

    def execute(self):
        pprint(self.content)
        return self.content



from openagi.llms.base import LLMBaseModel, LLMConfigModel
from openagi.utils.yamlParse import read_from_env
from typing import Any

from langchain_core.messages import HumanMessage

from openagi.exception import OpenAGIException
try:
   from langchain_anthropic import ChatAnthropic
except ImportError:
  raise OpenAGIException("Install Langchain Anthropic to use Claude LLM `pip install langchain-anthropic`")

class ChatAnthropicConfigModel(LLMConfigModel):
    """
    Configuration model Anthropic model. This provides opus, sonnet SOTA models
    """
    anthropic_api_key: str
    temperature: float = 0.5
    model_name: str = "claude-3-5-sonnet-20240620"

class ChatAnthropicModel(LLMBaseModel):
    """
    Define the Claude LLM from Anthropic using Langchain LLM integration
    """
    config: Any

    def load(self):
        """Initializes the ChatAnthropic instance with configurations."""
        self.llm = ChatAnthropic(
            model_name = self.config.model_name,
            api_key = self.config.anthropic_api_key,
            temperature = self.config.temperature
        )
        return self.llm

    def run(self, input_data: str):
        """
        Runs the Chat Anthropic model with the provided input text.
        Args:
            input_data: The input text to process.
        Returns:
            The response from Anthropic - Claude LLM.
        """

        if not self.llm:
            self.load()
        if not self.llm:
            raise ValueError("`llm` attribute not set.")
        
        message = HumanMessage(content=input_data)
        response = self.llm([message])
        return response.content

    @staticmethod
    def load_from_env_config() -> ChatAnthropicConfigModel:
        """Loads the ChatAnthropic configurations from a env file.

        Returns:
            An instance of ChatAnthropicConfigModel with loaded configurations.
        """
        return ChatAnthropicConfigModel(
            anthropic_api_key = read_from_env("ANTHROPIC_API_KEY", raise_exception=True),
            model_name = read_from_env("CLAUDE_MODEL_NAME",raise_exception=False),
            temperature = read_from_env("TEMPERATURE",raise_exception=False)
        )



import importlib
import json
import logging
import re
from textwrap import dedent
from typing import Dict, List, Optional, Tuple

from openagi.exception import OpenAGIException
from openagi.llms.base import LLMBaseModel


def force_json_output(resp_txt: str, llm):
    """
    Forces the output once the max iterations are reached.
    """
    #prompt = dedent(
    #    """
    #    Below is a JSON block. Please try to provide the output in the format shown below only
    #    ```json
    #        {"key": "value"}
    #    ```
    #    the contents between ```json and ``` will be extracted and passed to json.loads() in python to convert it to a dictionary. Make sure that it works when passed else you will be fined. If its already in the correct format, then you can return the same output in the expected output format.
    #    Input:
    #    {resp_txt}
    #    Output:
    #    """
    #).strip()

    prompt = dedent(
        """
        Your task is to process the input JSON and provide a valid JSON output. Follow these instructions carefully:
        1. The output must be enclosed in a code block using triple backticks and the 'json' language identifier, like this:
        ```json
         {"key": "value"}
        ```
        2. The JSON inside the code block must be valid and parseable by Python's json.loads() function.
        3. Ensure there are no extra spaces, newlines, or characters outside the JSON object within the code block.
        4. If the input is already in the correct format, reproduce it exactly in the output format specified above.
        5. Do not include any explanations, comments, or additional text in your response. The output needs be in JSON only. 
        6. Verify your output carefully before submitting. Incorrect responses will result in penalties.
        
        Input: {resp_txt}
        Output:
        """
    ).strip()

    prompt = prompt.replace("{resp_txt}", resp_txt)
    return llm.run(prompt)


def get_last_json(
    text: str, llm: Optional[LLMBaseModel] = None, max_iterations: int = 5
) -> Optional[Dict]:
    """
    Extracts the last block of text between ```json and ``` markers from a given string.

    Args:
        text (str): The string from which to extract the JSON block.
        llm (Optional[LLMBaseModel]): The language model instance to use for reformatting.
        max_iterations (int): Maximum number of iterations to try reformatting.

    Returns:
        dict or None: The last JSON block as a dictionary if found and parsed, otherwise None.
    """
    pattern = r"```json(.*?)```"
    matches = re.findall(pattern, text, flags=re.DOTALL)
    if matches:
        last_json = matches[-1].strip().replace("\n", "")
        try:
            return json.loads(last_json)
        except json.JSONDecodeError:
            logging.error("JSON not extracted. Trying again.", exc_info=True)
            pass

    if llm:
        for iteration in range(1, max_iterations + 1):
            logging.info(f"Iteration {iteration} to extract JSON from LLM output.")
            try:
                text = force_json_output(text, llm)
                matches = re.findall(pattern, text, flags=re.DOTALL)
                if matches:
                    last_json = matches[-1].strip().replace("\n", "")
                    json_resp = json.loads(last_json)
                    logging.info("JSON extracted successfully.")
                    return json_resp
            except json.JSONDecodeError:
                logging.error("JSON not extracted. Trying again.", exc_info=True)
                continue
            if iteration == max_iterations:
                raise OpenAGIException(
                    "The last output is not a valid JSON. Please check the output of the last action."
                )
    return None


def get_act_classes_from_json(json_data) -> List[Tuple[str, Optional[Dict]]]:
    """
    Extracts the Action class names and parameters from a JSON block.

    Args:
        json_data (List[Dict]): A list of dictionaries containing the class and parameter information.

    Returns:
        List[Tuple[type, Optional[Dict]]]: A list of tuples containing the Action class and its initialization parameters.
    """
    actions = []

    for item in json_data:
        # Extracting module and class name
        module_name = item["cls"]["module"]
        class_name = item["cls"]["kls"]

        # Dynamically import the module
        module = importlib.import_module(module_name)

        # Get the class from the module
        cls = getattr(module, class_name)

        # Extracting parameters for class initialization
        params = item["params"]

        # Storing the instance in the list
        actions.append((cls, params))

    return actions


def find_last_r_failure_content(text):
    """
    Finds the content of the last <r_failure> tag in the given text.

    Args:
        text (str): The text to search for the <r_failure> tag.

    Returns:
        str or None: The content of the last <r_failure> tag, or None if no matches are found.
    """
    pattern = r"<r_failure>(.*?)</r_failure>"
    matches = list(re.finditer(pattern, text, re.DOTALL))
    if matches:
        last_match = matches[-1]
        return last_match.group(1)
    else:
        return None


def extract_str_variables(template):
    """
    Extracts all variable names from a given template string.

    The function uses a regular expression to find all placeholders within curly braces in the template string, and returns a list of the extracted variable names.

    Args:
        template (str): The template string to extract variables from.

    Returns:
        list[str]: A list of variable names extracted from the template.
    """
    # This regular expression will find all placeholders within curly braces
    pattern = r"\{(\w+)\}"
    matches = re.findall(pattern, template)
    return matches



import logging
import nltk
from langchain_community.document_loaders import WebBaseLoader
from pydantic import Field
from sumy.nlp.stemmers import Stemmer
from sumy.nlp.tokenizers import Tokenizer
from sumy.parsers.plaintext import PlaintextParser
from sumy.summarizers.lsa import LsaSummarizer
from sumy.utils import get_stop_words

from openagi.actions.base import BaseAction
nltk.download("punkt")

class WebBaseContextTool(BaseAction):
    """
    Use this Action to extract actual context from a Webpage. The WebBaseContextTool class provides a way to load and optionally summarize the content of a webpage, returning the metadata and page content as a context string.
    If a url seems to be failing for more than once, ignore it and move forward.
    """

    link: str = Field(
        default_factory=str,
        description="Extract context for the Agents from the Web Search",
    )
    can_summarize: bool = Field(
        default=True,
        description="Indicates whether the action can summarize the content before returning. Uses ntlk & sumy.summarizers.lsa.LsaSummarizer. Defaults to true.",
    )

    def _get_summary(self, data):
        parser = PlaintextParser.from_string(data, Tokenizer("english"))
        summarizer = LsaSummarizer(Stemmer("english"))
        summarizer.stop_words = get_stop_words("english")

        summary = summarizer(parser.document, 6)
        summary_sentences = [str(sentence) for sentence in summary]
        return " ".join(summary_sentences)

    def execute(self):
        loader = WebBaseLoader(self.link)
        data = loader.load()
        metadata = data[0].metadata["title"]
        page_content = data[0].page_content
        if page_content:
            page_content = page_content.strip()
        if self.can_summarize:
            logging.info(f"Summarizing the page {self.link}...")
            page_content = self._get_summary(page_content)
        context = metadata + page_content
        return context



import os


def read_from_env(attr_name, raise_exception=False):
    attr_value = os.environ.get(attr_name)
    if not attr_value and raise_exception:
        raise ValueError(f"Unable to get config {attr_name}")
    return attr_value



import logging
from pathlib import Path
from typing import Dict, Optional

from pydantic import Field

from openagi.actions.base import BaseAction


class CreateFileAction(BaseAction):
    """
    Creates a new file with the specified content and directory structure.
    """

    filename: str = Field(..., description="Name of the file along with the directory.")
    parent_mkdir: bool = Field(
        default=True, description="Create parent directories of the file if not exist."
    )
    exist_ok: bool = Field(
        default=True,
        description="Do not raise error if any of the parent directories exists.",
    )
    file_content: str = Field(default="", description="String content of the file to insert")
    write_text_kargs: Optional[Dict] = Field(
        default=None, description="Kwargs to be passed to pathlib's write_text method"
    )

    def execute(self):
        output_file = Path(self.filename)
        print(f"Created file - {output_file.absolute()}")
        output_file.parent.mkdir(
            parents=self.parent_mkdir,
            exist_ok=self.exist_ok,
        )

        write_kwargs = {}
        if self.write_text_kargs:
            write_kwargs = {**write_kwargs}

        output_file.write_text(data=self.file_content, **write_kwargs)
        return self.file_content


class WriteFileAction(BaseAction):
    """
    Executes the action to write the provided content to a file at the specified path.
    """

    filename: str = Field(..., description="Name of the file along with the directory.")
    file_content: str = Field(default="", description="String content of the file to insert")
    file_mode: str = Field(
        default="w",
        description="File mode to open the file with while using python's open() func. Defaults to 'w'",
    )

    def execute(self):
        output_file = Path(self.filename)
        logging.info(f"Writing file - {output_file.absolute()}")
        with open(output_file.absolute(), self.file_mode) as f:
            f.write(self.file_content)
        return self.file_content


class ReadFileAction(BaseAction):
    """
    Reads the contents of a file specified by the `file_path` parameter.
    """

    file_path: str = Field(..., description="Name of the file along with the directory.")

    def execute(self):
        output_file = Path(self.file_path)
        logging.info(f"Reading file - {output_file.absolute()}")
        with open(output_file.absolute(), "r") as f:
            return f.read()



class OpenAGIException(Exception):
    ...


class ExecutionFailureException(Exception):
    """Task Execution Failed"""


class LLMResponseError(OpenAGIException):
    """No useful Response found"""



from uuid import uuid4
from openagi.llms.openai import OpenAIModel


def get_default_llm():
    config = OpenAIModel.load_from_env_config()
    return OpenAIModel(config=config)


def get_default_id():
    return uuid4().hex



import logging
from pathlib import Path
import re
from textwrap import dedent
from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel, Field, field_validator

from openagi.actions.utils import run_action
from openagi.exception import OpenAGIException
from openagi.llms.base import LLMBaseModel
from openagi.memory.memory import Memory
from openagi.prompts.worker_task_execution import WorkerAgentTaskExecution
from openagi.tasks.task import Task
from openagi.utils.extraction import get_act_classes_from_json, get_last_json
from openagi.utils.helper import get_default_id


class Worker(BaseModel):
    id: str = Field(default_factory=get_default_id)
    role: str = Field(description="Role of the worker.")
    instructions: Optional[str] = Field(description="Instructions the worker should follow.")
    llm: Optional[LLMBaseModel] = Field(
        description="LLM Model to be used.",
        default=None,
        exclude=True,
    )
    memory: Optional[Memory] = Field(
        default_factory=list,
        description="Memory to be used.",
        exclude=True,
    )
    actions: Optional[List[Any]] = Field(
        description="Actions that the Worker supports",
        default_factory=list,
    )
    max_iterations: int = Field(
        default=20,
        description="Maximum number of steps to achieve the objective.",
    )
    output_key: str = Field(
        default="final_output",
        description="Key to be used to store the output.",
    )
    force_output: bool = Field(
        default=True,
        description="If set to True, the output will be overwritten even if it exists.",
    )

    # Validate output_key. Should contain only alphabets and only underscore are allowed. Not alphanumeric
    @field_validator("output_key")
    @classmethod
    def validate_output_key(cls, v, values, **kwargs):
        if not re.match("^[a-zA-Z_]+$", v):
            raise ValueError(
                f"Output key should contain only alphabets and only underscore are allowed. Got {v}"
            )
        return v

    class Config:
        arbitrary_types_allowed = True

    def worker_doc(self):
        """Returns a dictionary containing information about the worker, including its ID, role, description, and the supported actions."""
        return {
            "worker_id": self.id,
            "role": self.role,
            "description": self.instructions,
            "supported_actions": [action.cls_doc() for action in self.actions],
        }

    def provoke_thought_obs(self, observation):
        thoughts = dedent(f"""Observation: {observation}""".strip())
        return thoughts

    def should_continue(self, llm_resp: str) -> Union[bool, Optional[Dict]]:
        output: Dict = get_last_json(llm_resp, llm=self.llm, max_iterations=self.max_iterations)
        output_key_exists = bool(output and output.get(self.output_key))
        return (not output_key_exists, output)

    def _force_output(
        self, llm_resp: str, all_thoughts_and_obs: List[str]
    ) -> Union[bool, Optional[str]]:
        """Force the output once the max iterations are reached."""
        prompt = (
            "\n".join(all_thoughts_and_obs)
            + "Based on the previous action and observation, give me the output."
        )
        output = self.llm.run(prompt)
        cont, final_output = self.should_continue(output)
        if cont:
            prompt = (
                "\n".join(all_thoughts_and_obs)
                + f"Based on the previous action and observation, give me the output. {final_output}"
            )
            output = self.llm.run(prompt)
            cont, final_output = self.should_continue(output)
        if cont:
            raise OpenAGIException(
                f"LLM did not produce the expected output after {self.max_iterations} iterations."
            )
        return (cont, final_output)

    def save_to_memory(self, task: Task):
        """Saves the output to the memory."""
        return self.memory.update_task(task)

    def execute_task(self, task: Task, context: Any = None) -> Any:
        """Executes the specified task."""
        logging.info(
            f"{'>'*20} Executing Task - {task.name}[{task.id}] with worker - {self.role}[{self.id}] {'<'*20}"
        )
        iteration = 1
        task_to_execute = f"{task.description}"
        worker_description = f"{self.role} - {self.instructions}"
        all_thoughts_and_obs = []

        logging.debug("Provoking initial thought observation...")
        initial_thought_provokes = self.provoke_thought_obs(None)
        te_vars = dict(
            task_to_execute=task_to_execute,
            worker_description=worker_description,
            supported_actions=[action.cls_doc() for action in self.actions],
            thought_provokes=initial_thought_provokes,
            output_key=self.output_key,
            context=context,
            max_iterations=self.max_iterations,
        )

        logging.debug("Generating base prompt...")
        base_prompt = WorkerAgentTaskExecution().from_template(te_vars)
        prompt = f"{base_prompt}\nThought:\nIteration: {iteration}\nActions:\n"

        logging.debug("Running LLM with prompt...")
        observations = self.llm.run(prompt)
        logging.info(f"LLM execution completed. Observations: {observations}")
        all_thoughts_and_obs.append(prompt)

        max_iters = self.max_iterations + 1
        while iteration < max_iters:
            logging.info(f"---- Iteration {iteration} ----")
            logging.debug("Checking if task should continue...")
            continue_flag, output = self.should_continue(observations)

            logging.debug("Extracting action from output...")
            action = output.get("action") if output else None
            if action:
                action = [action]

            # Save to memory
            if output:
                logging.debug("Saving task result and actions to memory...")
                task.result = observations
                task.actions = str([action.cls_doc() for action in self.actions])
                self.save_to_memory(task=task)

            if not continue_flag:
                logging.info(f"Task completed. Output: {output}")
                break

            if not action:
                logging.warning(f"No action found in the output: {output}")
                observations = f"Action: {action}\n{observations} Unable to extract action. Verify the output and try again."
                all_thoughts_and_obs.append(observations)
                iteration += 1
                continue

            if action:
                action_json = f"```json\n{output}\n```\n"
                try:
                    logging.debug("Getting action classes from JSON...")
                    actions = get_act_classes_from_json(action)
                    logging.info(
                        f"Extracted actions: {[act_cls.__name__ for act_cls, _ in actions]}"
                    )
                except KeyError as e:
                    if "cls" in e or "module" in e or "kls" in e:
                        observations = f"Action: {action_json}\n{observations}"
                        all_thoughts_and_obs.append(action_json)
                        all_thoughts_and_obs.append(observations)
                        iteration += 1
                        continue
                    else:
                        raise e

                for act_cls, params in actions:
                    params["memory"] = self.memory
                    params["llm"] = self.llm
                    try:
                        logging.debug(f"Running action: {act_cls.__name__}...")
                        res = run_action(action_cls=act_cls, **params)
                        logging.info(f"Action '{act_cls.__name__}' completed. Result: {res}")
                    except Exception as e:
                        if str(e).startswith("API KEY NOT FOUND"):
                            logging.error(e)
                            return {"error": "API KEY NOT FOUND. Please enter a valid API key before running this task."}, task
                        
                        observations = f"Action: {action_json}\n{observations}. {e} Try to fix the error and try again. Ignore if already tried more than twice"
                        all_thoughts_and_obs.append(action_json)
                        all_thoughts_and_obs.append(observations)
                        iteration += 1
                        continue

                    observation_prompt = f"Observation: {res}\n"
                    all_thoughts_and_obs.append(action_json)
                    all_thoughts_and_obs.append(observation_prompt)
                    observations = res

                logging.debug("Provoking thought observation...")
                thought_prompt = self.provoke_thought_obs(observations)
                all_thoughts_and_obs.append(f"\n{thought_prompt}\nActions:\n")

                prompt = f"{base_prompt}\n" + "\n".join(all_thoughts_and_obs)
                logging.debug(f"\nSTART:{'*' * 20}\n{prompt}\n{'*' * 20}:END")
                pth = Path(f"{self.memory.session_id}/logs/{task.name}-{iteration}.log")
                pth.parent.mkdir(parents=True, exist_ok=True)
                with open(pth, "w", encoding="utf-8") as f:
                    f.write(f"{prompt}\n")
                logging.debug("Running LLM with updated prompt...")
                observations = self.llm.run(prompt)
            iteration += 1
        else:
            if iteration == self.max_iterations:
                logging.info("---- Forcing Output ----")
                if self.force_output:
                    logging.debug("Forcing output...")
                    cont, final_output = self._force_output(observations, all_thoughts_and_obs)
                    if cont:
                        raise OpenAGIException(
                            f"LLM did not produce the expected output after {iteration} iterations for task {task.name}"
                        )
                    output = final_output
                    logging.debug("Saving final task result and actions to memory...")
                    task.result = observations
                    task.actions = str([action.cls_doc() for action in self.actions])
                    self.save_to_memory(task=task)
                else:
                    raise OpenAGIException(
                        f"LLM did not produce the expected output after {iteration} iterations for task {task.name}"
                    )

        logging.info(
            f"Task Execution Completed - {task.name} with worker - {self.role}[{self.id}] in {iteration} iterations"
        )
        return output, task



import logging
from typing import Any, Dict
from uuid import uuid4

from pydantic import BaseModel, Field

from openagi.storage.base import BaseStorage
from openagi.storage.chroma import ChromaStorage
from openagi.tasks.lists import TaskLists
from openagi.tasks.task import Task


class BaseMemory(BaseModel):
    session_id: str = Field(default_factory=lambda: uuid4().hex)
    storage: BaseStorage = Field(
        default_factory=lambda: ChromaStorage,
        description="Storage to be used for the Memory.",
        exclude=True,
    )

    def __init__(self, **data: Any):
        super().__init__(**data)
        self.storage = ChromaStorage.from_kwargs(collection_name=self.session_id)
        logging.info(f"Session ID initialized: {self.session_id}")

    def search(self, query: str, n_results: int = 10, **kwargs) -> Dict[str, Any]:
        """
        Search for similar tasks based on a query.

        :param query: The query string to search for.
        :param n_results: The number of results to return.
        :return: A dictionary of search results.
        """
        query_data = {
            "query_texts": query,
            "n_results": n_results,
            "include": ["metadatas", "documents"],
            **kwargs,
        }
        resp = self.storage.query_documents(**query_data)

        return resp["documents"]

    def display_memory(self) -> Dict[str, Any]:
        """
        Retrieve and display the current memory state from the database.

        :return: A dictionary of the current memory state.
        """
        result = self.storage.query_documents(self.session_id, n_results=2)
        return result or {}

    def save_task(self, task: Task) -> None:
        """
        Save execution details into Memory.

        :param task: The task to be saved.
        """
        metadata = self._create_metadata(task)
        self.storage.save_document(
            id=task.id,
            document=task.result,
            metadata=metadata,
        )
        logging.info(f"Task saved: {task.id}")

    def save_planned_tasks(self, tasks: TaskLists) -> None:
        """
        Save a list of planned tasks into Memory.

        :param tasks: The list of tasks to be saved.
        """
        for task in tasks:
            self.save_task(task)

    def update_task(self, task: Task) -> None:
        """
        Update a task in the Memory.

        :param task: The task to be updated.
        """
        metadata = self._create_metadata(task)
        self.storage.update_document(
            id=task.id,
            document=task.result,
            metadata=metadata,
        )
        logging.info(f"Task updated: {task.id}")

    def _create_metadata(self, task: Task) -> Dict[str, Any]:
        """
        Create metadata dictionary for a given task.

        :param task: The task for which to create metadata.
        :return: A dictionary of metadata.
        """
        return {
            "task_id": task.id,
            "session_id": self.session_id,
            "task_name": task.name,
            "task_description": task.description,
            "task_actions": task.actions,
        }



from openagi.exception import OpenAGIException
from openagi.llms.base import LLMBaseModel, LLMConfigModel
from openagi.utils.yamlParse import read_from_env

import logging
from typing import Any
from langchain_core.messages import HumanMessage

try:
   from langchain_mistralai import ChatMistralAI
except ImportError:
  raise OpenAGIException("Install langchain Mistral AI with cmd `pip install langchain_mistralai`")


class MistralConfigModel(LLMConfigModel):
    """Configuration model for Mistral."""

    mistral_api_key: str
    model_name: str = "mistral-large-latest"
    temperature: float = 0.1

class MistralModel(LLMBaseModel):
    """Mistral service implementation of the LLMBaseModel.

    This class implements the specific logic required to work with Mistral service.
    """

    config: Any

    def load(self):
        """Initializes the Mistral instance with configurations."""
        self.llm = ChatMistralAI(
           model = self.config.model_name,
           temperature = self.config.temperature,
           api_key = self.config.mistral_api_key
        )
        return self.llm

    def run(self, input_text: str):
        """Runs the Mistral model with the provided input text.

        Args:
            input_text: The input text to process.

        Returns:
            The response from Mistral service.
        """
        logging.info(f"Running LLM - {self.__class__.__name__}")
        if not self.llm:
            self.load()
        if not self.llm:
            raise ValueError("`llm` attribute not set.")
        message = HumanMessage(content=input_text)
        resp = self.llm([message])
        return resp.content

    @staticmethod
    def load_from_env_config() -> MistralConfigModel:
        """Loads the Mistral configurations from a YAML file.

        Returns:
            An instance of MistralConfigModel with loaded configurations.
        """
        return MistralConfigModel(
            mistral_api_key=read_from_env("MISTRAL_API_KEY", raise_exception=True),
        )


from openagi.llms.azure import AzureChatOpenAIModel
from openagi.agent import Admin
from openagi.memory import Memory
from openagi.worker import Worker
from openagi.planner.task_decomposer import TaskPlanner
from openagi.actions.base import BaseAction

import wikipedia
import os
import joblib
import requests
import re
import string
from tqdm import tqdm
from collections import Counter
from pydantic import Field, field_validator
import numpy as np

class WikiSearchAction(BaseAction):
  """
  Use this Action to get the information from Wikipedia Search
  """
  query: str = Field(
      default_factory=str,
      description="The search string. be simple"
    )

  def execute(self):
    search_res = wikipedia.search(self.query)
    if not search_res:
      return 'No results found.'
    article = wikipedia.page(search_res[0])
    return article.summary

def download_file(url, filename):
    """
    Download a file from a URL and save it locally.
    """
    response = requests.get(url)
    response.raise_for_status()
    with open(filename, "wb") as f:
        f.write(response.content)
    print(f"Downloaded {filename}")


def load_hotpot_qa_data(level):
    """
    Load HotpotQA data for a given level. If data doesn't exist, download it.
    """
    file_path = f"./data/{level}.joblib"
    data_url = (
        f"https://github.com/salesforce/BOLAA/raw/main/hotpotqa_run/data/{level}.joblib"
    )

    if not os.path.exists(file_path):
        print(f"{level} data not found, downloading...")
        # Create directory if it doesn't exist
        os.makedirs(os.path.dirname(file_path), exist_ok=True)
        download_file(data_url, file_path)
    # joblib requires python 3.10 or higher
    return joblib.load(file_path)

def normalize_answer(s):
    """
    Normalize answers for evaluation.
    """

    def remove_articles(text):
        return re.sub(r"\b(a|an|the)\b", " ", text)

    def white_space_fix(text):
        return " ".join(text.split())

    def remove_punc(text):
        exclude = set(string.punctuation)
        return "".join(ch for ch in text if ch not in exclude)

    def lower(text):
        return text.lower()

    return white_space_fix(remove_articles(remove_punc(lower(s))))


def f1_score(prediction, ground_truth):
    """
    Compute the F1 score between prediction and ground truth answers.
    """
    prediction_tokens = normalize_answer(prediction).split()
    ground_truth_tokens = normalize_answer(ground_truth).split()
    common = Counter(prediction_tokens) & Counter(ground_truth_tokens)
    num_same = sum(common.values())
    if num_same == 0:
        return 0, 0, 0
    precision = 1.0 * num_same / len(prediction_tokens)
    recall = 1.0 * num_same / len(ground_truth_tokens)
    f1 = (2 * precision * recall) / (precision + recall)
    return f1, precision, recall

def agent(query , llm):
    planner = TaskPlanner(autonomous=True)
    # for autonomous = False
    """
    planner = TaskPlanner(autonomous=False,human_intervene=False)
    
    researcher = Worker(
        role="Wikipedia Researcher",
        instructions="You are an intelligent Wikipedia agent. Your generation should follow the example format. Finish the task if you find the answer. And you answer should be simple and straighforward. DO NOT repeat your actions.",
        actions=[WikiSearchAction],
        max_iterations = 3,
    )"""
    admin = Admin(
        planner=planner,
        memory=Memory(),
        actions=[WikiSearchAction],
        llm=llm,
    )
    # admin.assign_workers([researcher]) #un-comment this for Autonomous = False

    res = admin.run(
        query=query,
        description="Provide answer for the query.You should decompose your task into executable actions.",
    )
    return res

def run_agent(level = 'easy'):
    os.environ["AZURE_BASE_URL"] = ""
    os.environ["AZURE_DEPLOYMENT_NAME"] = ""
    os.environ["AZURE_MODEL_NAME"]="gpt4"
    os.environ["AZURE_OPENAI_API_VERSION"]=""
    os.environ["AZURE_OPENAI_API_KEY"]=  ""
    config = AzureChatOpenAIModel.load_from_env_config()
    llm = AzureChatOpenAIModel(config=config)

    hotpot_data = load_hotpot_qa_data(level)
    hotpot_data = hotpot_data.reset_index(drop=True)
    task_instructions = [
        (row["question"], row["answer"]) for _, row in hotpot_data.iterrows()
    ]

    f1_list = []
    correct = 0
    results = {}

    for task , answer in tqdm(task_instructions[0:30]):
        response = agent(task , llm)
        f1 , _ ,_ = f1_score(response,answer)
        f1_list.append(f1)
        correct += int(response == answer)

        avg_f1 = np.mean(f1_list)
        acc = correct / len(task_instructions[0:30])
    return avg_f1, acc

# levels are 'easy' , 'medium' , 'hard'
f1 , acc = run_agent(level='easy')
print(f"F1 score : {f1} , Accuracy : {acc}")


from typing import Any
from langchain_core.messages import HumanMessage
from langchain_community.llms import HuggingFaceHub
from openagi.llms.base import LLMBaseModel, LLMConfigModel
from openagi.utils.yamlParse import read_from_env

class HuggingFaceConfigModel(LLMConfigModel):
    """Configuration model for Hugging Face."""
    api_token: str
    model_name: str = "huggingfaceh4/zephyr-7b-beta"
    temperature: float = 0.1
    max_new_tokens: int = 512

class HuggingFaceModel(LLMBaseModel):
    """Hugging Face service implementation of the LLMBaseModel.

    This class implements the specific logic required to work with Hugging Face service.
    """

    config: Any

    def load(self):
        """Initializes the GroqModel instance with configurations."""
        self.llm = HuggingFaceHub(
            huggingfacehub_api_token = self.config.api_token,
            repo_id= self.config.model_name, 
            model_kwargs={"temperature": self.config.temperature,
                          "max_new_tokens":self.config.max_new_tokens,
                          "repetition_penalty":1.2}
        )
        return self.llm
    
    def run(self, input_data: str):
        """Runs the HuggingFace model with the provided input text.

        Args:
            input_data: The input text to process.

        Returns:
            The response from HuggingFace model.
        """
        if not self.llm:
            self.load()
        if not self.llm:
            raise ValueError("`llm` attribute not set.")
        message = HumanMessage(content=input_data)
        resp = self.llm([message])
        return resp.content

    @staticmethod
    def load_from_env_config() -> HuggingFaceConfigModel:
        """Loads the Hugging Face configurations from a YAML file.

        Returns:
            An instance of HuggingFaceConfigModel with loaded configurations.
        """
        return HuggingFaceConfigModel(
            api_token = read_from_env("HUGGINGFACE_ACCESS_TOKEN",raise_exception=True),
            model_name=read_from_env("HUGGINGFACE_MODEL", raise_exception=True),
            temperature=read_from_env("TEMPERATURE",raise_exception=True),
            max_new_tokens= read_from_env("MAX_NEW_TOKENS",raise_exception=True)
        )


import base64
import os
from typing import Dict, List
from openagi.exception import OpenAGIException
import requests
from langchain_community.document_loaders.github import GithubFileLoader
from pydantic import Field
from openagi.actions.base import BaseAction

import warnings
warnings.filterwarnings("ignore")

class OpenAGIGithubFileLoader(GithubFileLoader):
    def get_file_paths(self) -> List[Dict]:
        base_url = (
            f"{self.github_api_url}/repos/{self.repo}/git/trees/" f"{self.branch}?recursive=1"
        )
        response = requests.get(base_url, headers=self.headers)
        response.raise_for_status()
        all_files = response.json()["tree"]
        
        """ one element in all_files
        {
            'path': '.github', 
            'mode': '040000', 
            'type': 'tree', 
            'sha': '89a2ae046e8b59eb96531f123c0c6d4913885df1', 
            'url': 'https://github.com/api/v3/repos/shufanhao/langchain/git/trees/89a2ae046e8b59eb96531f123c0c6d4913885dxxx'
        }
        """
        required_files = [
            f for f in all_files if not (self.file_filter and not self.file_filter(f["path"]))
        ]
        return required_files

    def get_file_content_by_path(self, path: str) -> str:
        base_url = f"{self.github_api_url}/repos/{self.repo}/contents/{path}"
        #print(base_url)
        response = requests.get(base_url, headers=self.headers)
        response.raise_for_status()

        content_encoded = response.json()["content"]
        return base64.b64decode(content_encoded).decode("utf-8")
        

class GitHubFileLoadAction(BaseAction):
    """
    #Use this Action to extract specific extension files from GitHub.
    """

    repo: str = Field(
        default_factory=str,
        description="Repository name- Format: username/repo e.g., aiplanethub/openagi",
    )
    directory:str = Field(
        default_factory=str,
        description="File directory that contains the supporting files i.e., src/openagi/llms",
    )
    extension: str = Field(
        default_factory = ".txt",
        description="File extension to extract the data from. eg: `.py`, `.md`",
    )


    def execute(self):
        access_token = os.environ.get("GITHUB_ACCESS_TOKEN")

        loader = OpenAGIGithubFileLoader(
            repo=self.repo,
            access_token=access_token,
            github_api_url="https://api.github.com",
            branch="main",
            file_filter=lambda files: files.startswith(self.directory) and files.endswith(self.extension),
        )

        data = loader.load()
        response = []
        for doc in data:
            response.append(f"{doc.page_content}\nMetadata{doc.metadata}")

        return "\n\n".join(response)



import logging
from pydantic import Field
from openagi.exception import OpenAGIException
from openagi.actions.base import BaseAction

try:
   from unstructured.partition.pdf import partition_pdf
   from unstructured.chunking.title import chunk_by_title
except ImportError:
  raise OpenAGIException("Install Unstructured with cmd `pip install 'unstructured[all-docs]'`")


class UnstructuredPdfLoaderAction(BaseAction):
    """
    Use this Action to extract content from PDFs including metadata.
    Returns a list of dictionary with keys 'type', 'element_id', 'text', 'metadata'.
    """

    file_path: str = Field(
        default_factory=str,
        description="File or pdf file url from which content is extracted.",
    )

    def execute(self):
        logging.info(f"Reading file {self.file_path}")

        elements = partition_pdf(self.file_path, extract_images_in_pdf=True)

        chunks = chunk_by_title(elements)

        dict_elements = []
        for element in chunks:
            dict_elements.append(element.to_dict())

        with open("ele.txt", "w") as f:
            f.write(str(dict_elements))

        return str(dict_elements)



from abc import abstractmethod
from typing import Any

from pydantic import BaseModel


class LLMConfigModel(BaseModel):
    """Base configuration model for all LLMs.

    This class can be extended to include more fields specific to certain LLMs.
    """

    class Config:
        protected_namespaces = ()

    pass  # Common fields could be defined here, if any.


class LLMBaseModel(BaseModel):
    """Abstract base class for language learning models.

    Attributes:
        config: An instance of LLMConfigModel containing configuration.
        llm: Placeholder for the actual LLM instance, to be defined in subclasses.
    """

    config: Any
    llm: Any = None

    @abstractmethod
    def load(self):
        """Initializes the LLM instance with configurations."""
        pass

    @abstractmethod
    def run(self, input_data: Any):
        """Interacts with the LLM service using the provided input.

        Args:
            input_data: The input to process by the LLM. The format can vary.

        Returns:
            The result from processing the input data through the LLM.
        """
        pass

    @staticmethod
    @abstractmethod
    def load_from_env_config():
        """Loads configuration values from a YAML file."""
        pass



from pydantic import BaseModel, ConfigDict, Field


class BaseStorage(BaseModel):
    """Base Storage class to be inherited by other storages, providing basic functionality and structure."""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    name: str = Field(title="BaseStorage", description="Name of the Storage.")

    def save_document(self):
        """Save documents to the with metadata."""
        raise NotImplementedError("Subclasses must implement this method.")

    def update_document(self):
        raise NotImplementedError("Subclasses must implement this method.")

    def delete_document(self):
        raise NotImplementedError("Subclasses must implement this method.")

    def query_documents(self):
        raise NotImplementedError("Subclasses must implement this method.")

    @classmethod
    def from_kwargs(cls, **kwargs):
        raise NotImplementedError("Subclasses must implement this method.")



from typing import Any
from langchain_core.messages import HumanMessage
from langchain_openai import AzureChatOpenAI  # Assuming this import is correct

from openagi.llms.base import LLMBaseModel, LLMConfigModel
from openagi.utils.yamlParse import read_from_env


class AzureChatConfigModel(LLMConfigModel):
    """Configuration model for Azure Chat OpenAI."""

    base_url: str
    deployment_name: str
    model_name: str
    openai_api_version: str
    api_key: str


class AzureChatOpenAIModel(LLMBaseModel):
    """Azure's OpenAI service implementation of the LLMBaseModel.

    This class implements the specific logic required to work with Azure's OpenAI service.
    """

    config: Any

    def load(self):
        """Initializes the AzureChatOpenAI instance with configurations."""
        self.llm = AzureChatOpenAI(
            azure_deployment=self.config.deployment_name,
            model_name=self.config.model_name,
            openai_api_version=self.config.openai_api_version,
            openai_api_key=self.config.api_key,
            azure_endpoint=self.config.base_url,
        )
        return self.llm

    def run(self, input_data: str):
        """Runs the Azure Chat OpenAI model with the provided input text.

        Args:
            input_data: The input text to process.

        Returns:
            The response from Azure's OpenAI service.
        """
        if not self.llm:
            self.load()
        if not self.llm:
            raise ValueError("`llm` attribute not set.")
        message = HumanMessage(content=input_data)
        resp = self.llm([message])
        return resp.content

    @staticmethod
    def load_from_env_config() -> AzureChatConfigModel:
        """Loads the AzureChatOpenAI configurations from a YAML file.

        Returns:
            An instance of AzureChatConfigModel with loaded configurations.
        """
        return AzureChatConfigModel(
            base_url=read_from_env("AZURE_BASE_URL", raise_exception=True),
            deployment_name=read_from_env("AZURE_DEPLOYMENT_NAME", raise_exception=True),
            model_name=read_from_env("AZURE_MODEL_NAME", raise_exception=True),
            openai_api_version=read_from_env("AZURE_OPENAI_API_VERSION", raise_exception=True),
            api_key=read_from_env("AZURE_OPENAI_API_KEY", raise_exception=True),
        )



from openagi.actions.base import BaseAction
from pydantic import Field
from openagi.exception import OpenAGIException
from typing import ClassVar, Dict, Any

try:
    from exa_py import Exa
except ImportError:
    raise OpenAGIException("Install Exa Py with cmd `pip install exa_py`")

class ConfigurableAction(BaseAction):
    config: ClassVar[Dict[str, Any]] = {}

    @classmethod
    def set_config(cls, *args, **kwargs):
        if args:
            if len(args) == 1 and isinstance(args[0], dict):
                cls.config.update(args[0])
            else:
                raise ValueError("If using positional arguments, a single dictionary must be provided.")
        cls.config.update(kwargs)

    @classmethod
    def get_config(cls, key: str, default: Any = None) -> Any:
        return cls.config.get(key, default)

class ExaSearch(ConfigurableAction):
    """
    Exa Search is a tool used when user needs to ask the question in terms of query to get response 
    """
    query: str = Field(..., description="User query or question")

    def execute(self):
        api_key = self.get_config('api_key')
        if not api_key:
            raise OpenAGIException("API KEY NOT FOUND. Use ExaSearch.set_config(api_key='your_key') to set the API key.")

        exa = Exa(api_key=api_key)
        results = exa.search_and_contents(
            self.query,
            text={"max_characters": 512},
        )
        
        content = ""
        for idx in results.results:
            content += idx.text.strip()

        content = content.replace("<|endoftext|>", "")
        content = content.replace("NaN", "")
        return content


from openagi.actions.base import BaseAction
from langchain_community.document_loaders import TextLoader
from langchain_community.document_loaders.csv_loader import CSVLoader
from pydantic import Field


class DocumentLoader(BaseAction):
    """Use this Action to extract content from documents"""

    file_path: str = Field(
        default_factory=str,
        description="File from which content is extracted",
    )

    def text_loader(self):
        loader = TextLoader(file_path=self.file_path)
        data = loader.load()
        page_content = data[0].page_content
        meta_data = data[0].metadata["source"]
        context = meta_data + " " + page_content
        return context

    def csv_loader(self):
        content = ""
        loader = CSVLoader(file_path=self.file_path)
        data = loader.load()

        for i in range(len(data)):
            row_content = data[i].page_content
            row_no = data[i].metadata["row"]
            content += "row_no" + " " + str(row_no) + ": " + str(row_content)
        return content

    def execute(self):
        if self.file_path.endswith(".txt"):
            context = self.text_loader()
        elif self.file_path.endswith(".csv"):
            context = self.csv_loader()
        return context



from typing import Any

from pydantic import Field

from openagi.actions.base import BaseAction


class FormatterAction(BaseAction):
    """Content Formatter Action"""

    content: Any = Field(..., description="Data/Content to be formatted.")
    format_type: str = Field(
        default="markdown",
        description="Type to which the content will be formatted to. It will be modified to the supported formats and returned. Supported Formats - markdown/plan-text",
    )

    def execute(self):
        return self.llm.run(
            f"Format and return the below response in {self.format_type} format without removing any content. You can rephrase if required.\n{self.content}"
        )



import logging
from enum import Enum
from textwrap import dedent
from typing import Any, Dict, List, Optional, Union
from pydantic import BaseModel, Field, field_validator

from openagi.actions.base import BaseAction
from openagi.actions.compressor import SummarizerAction
from openagi.actions.formatter import FormatterAction
from openagi.actions.obs_rag import MemoryRagAction
from openagi.actions.utils import run_action
from openagi.exception import OpenAGIException
from openagi.llms.azure import LLMBaseModel
from openagi.memory.memory import Memory
from openagi.planner.task_decomposer import BasePlanner, TaskPlanner
from openagi.prompts.worker_task_execution import WorkerAgentTaskExecution
from openagi.tasks.lists import TaskLists
from openagi.utils.extraction import (
    find_last_r_failure_content,
    get_act_classes_from_json,
    get_last_json,
)
from openagi.utils.helper import get_default_llm
from openagi.utils.tool_list import get_tool_list
from openagi.worker import Worker

class OutputFormat(str, Enum):
    markdown = "markdown"
    raw_text = "raw_text"


class Admin(BaseModel):
    planner: Optional[BasePlanner] = Field(
        description="Type of planner to use for task decomposition.",
        default=None,
    )
    llm: Optional[LLMBaseModel] = Field(
        description="LLM Model to be used.",
        default=None,
    )
    memory: Optional[Memory] = Field(
        default_factory=list, description="Memory to be used.", exclude=True
    )
    actions: Optional[List[Any]] = Field(
        description="Actions that the Agent supports", default_factory=list
    )
    max_iterations: int = Field(
        default=20, description="Maximum number of steps to achieve the objective."
    )
    output_format: OutputFormat = Field(
        default=OutputFormat.markdown,
        description="Format to be converted the result while returning.",
    )
    workers: List[Worker] = Field(
        default_factory=list,
        description="List of workers managed by the Admin agent.",
    )
    summarize_task_context: bool = Field(
        default=True,
        description="If set to True, the task context will be summarized and passed to the next task else the task context will be passed as is.",
    )
    output_key: str = Field(
        default="final_output",
        description="Key to be used to store the output.",
    )

    def model_post_init(self, __context: Any) -> None:
        model = super().model_post_init(__context)

        if not self.llm:
            self.llm = get_default_llm()

        if not self.planner:
            self.planner = TaskPlanner(workers=self.workers)

        if not self.memory:
            self.memory = Memory()

        self.actions = self.actions or []

        default_actions = [MemoryRagAction]
        self.actions.extend(default_actions)

        return model

    @field_validator("actions")
    @classmethod
    def actions_validator(cls, act_clss):
        for act_cls in act_clss:
            if not issubclass(act_cls, BaseAction):
                raise ValueError(f"{act_cls} is not a subclass of BaseAction")
        return act_clss

    def assign_workers(self, workers: List[Worker]):
        if workers:
            for worker in workers:
                if not getattr(worker, "llm", False):
                    setattr(worker, "llm", self.llm)
                if not getattr(worker, "memory", False):
                    setattr(worker, "memory", self.memory)

        if not self.workers:
            self.workers = workers
        else:
            self.workers.extend(workers)

    def run_planner(self, query: str, descripton: str):
        if self.planner:
            if not getattr(self.planner, "llm", False):
                setattr(self.planner, "llm", self.llm)

            setattr(self.planner, "workers", self.workers)

        logging.info("Thinking...")
        actions_dict: List[BaseAction] = []

        for act in self.actions:
            actions_dict.append(act.cls_doc())

        workers_dict = []
        for worker in self.workers:
            workers_dict.append(worker.worker_doc())

        return self.planner.plan(
            query=query,
            description=descripton,
            supported_actions=actions_dict,
            supported_workers=workers_dict,
        )

    def _generate_tasks_list(self, planned_tasks):
        task_lists = TaskLists()
        task_lists.add_tasks(tasks=planned_tasks)
        logging.debug(f"Created {task_lists.get_tasks_queue().qsize()} Tasks.")
        return task_lists

    def get_previous_task_contexts(self, task_lists: TaskLists):
        task_summaries = []
        logging.info("Retrieving completed task contexts...")
        t_list = task_lists.completed_tasks.queue
        for indx, task in enumerate(t_list):
            memory = run_action(
                action_cls=MemoryRagAction,
                task=task,
                llm=self.llm,
                memory=self.memory,
                query=task.id,
            )
            if memory and self.summarize_task_context:
                params = {
                    "past_messages": memory,
                    "llm": self.llm,
                    "memory": self.memory,
                    "instructions": "Include summary of all the thoughts, but include all the relevant points from the observations without missing any.",
                }
                memory = run_action(action_cls=SummarizerAction, **params)
                if not memory:
                    raise Exception("No memory returned after summarization.")
            task_summaries.append(f"\n{indx+1}. {task.name} - {task.description}\n{memory}")
        else:
            logging.warning("No Tasks to summarize.")
        if task_summaries:
            return "\n".join(task_summaries).strip()
        return "None"

    def _get_worker_by_id(self, worker_id: str):
        for worker in self.workers:
            if worker.id == worker_id:
                return worker
        raise ValueError(f"Worker with id {worker_id} not found.")

    def worker_task_execution(self, query: str, description: str, task_lists: TaskLists):
        res = None

        while not task_lists.all_tasks_completed:
            cur_task = task_lists.get_next_unprocessed_task()
            worker = self._get_worker_by_id(cur_task.worker_id)
            res, task = worker.execute_task(
                cur_task,
                context=self.get_previous_task_contexts(task_lists=task_lists),
            )
            self.memory.update_task(task)
            task_lists.add_completed_tasks(task)

        logging.info("Finished Execution...")

        if self.output_format == OutputFormat.markdown and res:
            logging.info("Output Formatting...")
            output_formatter = FormatterAction(
                content=res,
                format_type=OutputFormat.markdown,
                llm=self.llm,
                memory=self.memory,
            )
            res = output_formatter.execute()
        logging.debug(f"Execution Completed for Session ID - {self.memory.session_id}")
        return res

    def _provoke_thought_obs(self, observation):
        thoughts = dedent(f"""Observation: {observation}""".strip())
        return thoughts

    def _should_continue(self, llm_resp: str) -> Union[bool, Optional[Dict]]:
        output: Dict = get_last_json(llm_resp, llm=self.llm, max_iterations=self.max_iterations)
        output_key_exists = bool(output and output.get(self.output_key))
        return (not output_key_exists, output)

    def _force_output(
        self, llm_resp: str, all_thoughts_and_obs: List[str]
    ) -> Union[bool, Optional[str]]:
        """Force the output once the max iterations are reached."""
        prompt = (
            "\n".join(all_thoughts_and_obs)
            + "Based on the previous action and observation, give me the output."
        )
        output = self.llm.run(prompt)
        cont, final_output = self._should_continue(output)
        if cont:
            prompt = (
                "\n".join(all_thoughts_and_obs)
                + f"Based on the previous action and observation, give me the output. {final_output}"
            )
            output = self.llm.run(prompt)
            cont, final_output = self._should_continue(output)
        if cont:
            raise OpenAGIException(
                f"LLM did not produce the expected output after {self.max_iterations} iterations."
            )
        return (cont, final_output)

    def auto_workers_assignment(self, query: str, description: str, task_lists: TaskLists):
        """
        Autonomously generates the Workers with the

        Args:
            query (str): The query to be processed.
            description (str): A description of the task.
            task_lists (TaskLists): The task lists to be processed.

        Returns:
            str: JSON of the list of Workers that needs to be executed
        """

        workers = []
        tools_list = get_tool_list()
        
        for action in self.actions:
            tools_list.append(action)

        worker_dict = {}
        main_task_list = TaskLists()
        while not task_lists.all_tasks_completed:
            cur_task = task_lists.get_next_unprocessed_task()
            print(cur_task)
            logging.info(f"**** Executing Task - {cur_task.name} [{cur_task.id}] ****")

            worker_config = cur_task.worker_config

            worker_instance = None
            if worker_config["role"] not in worker_dict:
                worker_instance = Worker(
                    role=worker_config["role"],
                    instructions=worker_config["instructions"],
                    llm=self.llm,
                    actions=self.get_supported_actions_for_worker(
                        worker_config["supported_actions"],tools_list
                    ),
                )
                worker_dict[worker_config["role"]] = worker_instance
            else:
                worker_instance = worker_dict[worker_config["role"]]
            workers.append(worker_instance)
            cur_task.worker_id = worker_instance.id
            main_task_list.add_task(cur_task)

        task_lists = main_task_list
        self.assign_workers(workers=workers)

        if self.workers:
            return self.worker_task_execution(
                query=query,
                description=description,
                task_lists=task_lists,
            )

    def single_agent_execution(self, query: str, description: str, task_lists: TaskLists):
        """
        Executes a single agent's tasks for the given query and description, updating the task lists and memory as necessary.

        Args:
            query (str): The query to be processed.
            description (str): A description of the task.
            task_lists (TaskLists): The task lists to be processed.

        Returns:
            str: The final result of the task execution.
        """
        all_thoughts_and_obs = []
        output = None
        previous_task_context = None

        while not task_lists.all_tasks_completed:
            iteration = 1
            max_iterations = self.max_iterations

            cur_task = task_lists.get_next_unprocessed_task()
            logging.info(f"**** Executing Task - {cur_task.name} [{cur_task.id}] ****")

            task_to_execute = f"{cur_task.name}. {cur_task.description}"
            agent_description = "Task executor"

            logging.debug("Provoking initial thought observation...")
            initial_thought_provokes = self._provoke_thought_obs(None)

            te_vars = dict(
                task_to_execute=task_to_execute,
                worker_description=agent_description,
                supported_actions=[action.cls_doc() for action in self.actions],
                thought_provokes=initial_thought_provokes,
                output_key=self.output_key,
                context=previous_task_context,
                max_iterations=max_iterations,
            )

            logging.debug("Generating base prompt...")
            base_prompt = WorkerAgentTaskExecution().from_template(te_vars)
            prompt = f"{base_prompt}\nThought:\nIteration: {iteration}\nActions:\n"

            logging.debug("Running LLM with prompt...")
            observations = self.llm.run(prompt)
            logging.info(f"LLM execution completed. Observations: {observations}")
            all_thoughts_and_obs.append(prompt)

            while iteration < max_iterations:
                logging.info(f"---- Iteration {iteration} ----")
                continue_flag, output = self._should_continue(observations)

                if not continue_flag:
                    logging.info(f"Task completed. Output: {output}")
                    break

                resp_json = get_last_json(observations)

                output = resp_json.get(self.output_key) if resp_json else None
                if output:
                    cur_task.result = output
                    cur_task.actions = te_vars["supported_actions"]
                    self.memory.update_task(cur_task)

                action_json = resp_json.get("action") if resp_json else None

                if action_json and not isinstance(action_json, list):
                    action_json = [action_json]

                if not action_json:
                    logging.warning(f"No action found in the output: {output}")
                    observations = f"Action: {action_json}\n{observations} Unable to extract action. Verify the output and try again."
                    all_thoughts_and_obs.append(observations)
                    iteration += 1
                elif action_json:
                    actions = get_act_classes_from_json(action_json)

                    for act_cls, params in actions:
                        params["previous_action"] = None  # Modify as needed
                        params["llm"] = self.llm
                        params["memory"] = self.memory
                        try:
                            logging.debug(f"Running action: {act_cls.__name__}...")
                            res = run_action(action_cls=act_cls, **params)
                            logging.info(f"Action '{act_cls.__name__}' completed. Result: {res}")
                        except Exception as e:
                            logging.error(f"Error running action: {e}")
                            observations = f"Action: {action_json}\n{observations}. {e} Try to fix the error and try again. Ignore if already tried more than twice"
                            all_thoughts_and_obs.append(observations)
                            iteration += 1
                            continue

                        observation_prompt = f"Observation: {res}\n"
                        all_thoughts_and_obs.append(observation_prompt)
                        observations = res

                    logging.debug("Provoking thought observation...")
                    thought_prompt = self._provoke_thought_obs(observations)
                    all_thoughts_and_obs.append(f"\n{thought_prompt}\nActions:\n")

                    prompt = f"{base_prompt}\n" + "\n".join(all_thoughts_and_obs)
                    logging.debug(f"\nSTART:{'*' * 20}\n{prompt}\n{'*' * 20}:END")
                    logging.debug("Running LLM with updated prompt...")
                    observations = self.llm.run(prompt)
                    iteration += 1
            else:
                if iteration == max_iterations:
                    logging.info("---- Forcing Output ----")
                    cont, final_output = self._force_output(observations, all_thoughts_and_obs)
                    if cont:
                        raise OpenAGIException(
                            f"LLM did not produce the expected output after {iteration} iterations for task {cur_task.name}"
                        )
                    output = final_output
                    cur_task.result = output
                    cur_task.actions = te_vars["supported_actions"]
                    self.memory.update_task(cur_task)
                    task_lists.add_completed_tasks(cur_task)

            previous_task_context = self.get_previous_task_contexts(task_lists)
            task_lists.add_completed_tasks(cur_task)

        logging.info("Finished Execution...")

        if self.output_format == OutputFormat.markdown:
            logging.info("Output Formatting...")
            output_formatter = FormatterAction(
                content=output,
                format_type=OutputFormat.markdown,
                llm=self.llm,
                memory=self.memory,
            )
            output = output_formatter.execute()

        logging.debug(f"Execution Completed for Session ID - {self.memory.session_id}")
        return output

    def run(self, query: str, description: str,planned_tasks: Optional[List[Dict]] = None):
        logging.info("Running Admin Agent...")
        logging.info(f"SessionID - {self.memory.session_id}")

        if planned_tasks is None:
            planned_tasks = self.run_planner(query=query, descripton=description)

        logging.info("Tasks Planned...")
        logging.debug(f"{planned_tasks=}")

        task_lists: TaskLists = self._generate_tasks_list(planned_tasks=planned_tasks)

        self.memory.save_planned_tasks(tasks=list(task_lists.tasks.queue))

        if self.planner.autonomous:
            return self.auto_workers_assignment(
                query=query, description=description, task_lists=task_lists
            )
        else:
            if self.workers:
                return self.worker_task_execution(
                    query=query,
                    description=description,
                    task_lists=task_lists,
                )
            else:
                return self.single_agent_execution(
                    query=query, description=description, task_lists=task_lists
                )

    def _can_task_execute(self, llm_resp: str) -> Union[bool, Optional[str]]:
        content: str = find_last_r_failure_content(text=llm_resp)
        if content:
            return False, content
        return True, content

    def get_supported_actions_for_worker(self, actions_list: List[str],tool_list: List[str]):
        """
        This function takes a list of action names (strings) and returns a list of class objects
        from the modules within the 'tools' folder that match these action names and inherit from BaseAction.

        :param actions_list: List of action names as strings.
        :return: List of matching class objects.
        """
        matching_classes = []
        #tool_list = get_tool_list()
        # Iterate through all modules in the tools package
        for action in tool_list:
            if action.__name__ in actions_list:
                matching_classes.append(action)

        return matching_classes



from openagi.prompts.base import BasePrompt

TASK_CLARIFICATION_PROMPT = """
As an AI clarity assistant for OpenAGI, your job is to ensure tasks are completely unambiguous. Analyze the given Task_Objectives, Task_Descriptions, and the conversation history. Identify any unclear or missing information.

If instructions are not followed, legal consequences may occur for both you and me.

Instructions:
1. Examine the task for ambiguities or missing crucial details.
2. Atleast ask one clarify question in beginning, even if the task is clear
3. If unclear points exist, formulate a single, specific question addressing the most critical ambiguity.
4. Do not repeat questions or ask about information already provided.
5. If the task is clear, no new questions are needed, be smart and return an empty string.
6. If the last human response indicates unwillingness to clarify (e.g., "I don't know", "No more questions", "That's all I can say"), return an empty string.

Input:
- Task_Objectives: {objective}
- Task_Descriptions: {task_descriptions}
- Conversation_History: {chat_history}

Output Format

Always return a JSON object, enclosed in triple backticks:
If clarification is needed:
```json
{
    "question": "<Your clarification question here>"
}
```
If NO clarification is needed or Task is clear:
```json
{
    "question": ""
}
"""

class TaskClarifier(BasePrompt):
    base_prompt: str = TASK_CLARIFICATION_PROMPT



import logging
from pydantic import Field
from openagi.actions.base import BaseAction

class MemoryRagAction(BaseAction):
    """Action class to get all the results from the previous tasks for the current objetive.
    This action is responsible to reading and not writing. Writing is done by default for every task.
    """

    query: str = Field(
        ...,
        description="Query, a string, to run to retrieve the data from the results of previous tasks. Returns an Array of the results.",
    )
    max_results: int = Field(
        default=10,
        description="Max results to be used by querying the memory Defaults to integer 10.",
    )

    def execute(self):
        resp = self.memory.search(query=self.query, n_results=self.max_results or 10)
        logging.debug(f"Retreived MEMORY DATA  -  {resp}")
        return resp



from textwrap import dedent
from openagi.prompts.base import BasePrompt

single_agent_task_creation = """
You are a task-creator AI for OpenAGI. Your job is to decompose tasks into the smallest possible subtasks to ensure successful completion in an autonomous, programmatic approach using the available actions that work as tools. Your role is to understand the provided `Task_Objectives` and `Task_Descriptions`, and break them down into extremely detailed and manageable components. Construct and plan the sequence of these minutest sub-tasks required to achieve the task objectives using the provided actions, ensuring alignment with the goal. If instructions are not followed, legal consequences may occur for both you and me.

**Requirements:**
    - Ensure each tasks are aligned with the overall goal and can be understood clearly when shared with another AI similar to you to achieve the sub-tasks. Each task will be completely run by another AI where they will be getting the results from the previous task, without knowledge of how it was executed.
    - Understand the parameters of each supported action when using them.
    - Only One Action per task. Ensure tasks are decomposed in the same manner.
    - Please ensure each task meets the criteria above and refine as necessary to maintain clarity and alignment with the overall objectives.
    - If no tasks, return the reasons why no tasks can be created.

**Inputs:**
    - Task_Objectives: {objective}
    - Task_Descriptions: {task_descriptions}
    - SUPPORTED_ACTIONS: {supported_actions}

**Output Format:**
Return the tasks in JSON format with the keys "task_name" and "description". Ensure the JSON format is suitable for utilization with JSON.parse(), enclosed in triple backticks.
```json
[
    {
        "task_name": "<name of the task of type string>",
        "description": "<description of the task of type string>"
    }
]
```

**Notes:**
    - You do not need to create tasks for storing the results, as results will be stored automatically after executing each task. You can retrieve previous task results using MemoryRagAction.

**Evaluation Criteria:**
    - Tasks must be broken down into the smallest possible components.
    - Each task must be clear and executable by an AI agent.
    - Tasks must follow a logical sequence to achieve the overall objective.
    - Ensure alignment with the provided actions and goals.
"""


worker_task_creation = dedent(
"""
You are a task-creator AI for OpenAGI. Your job is to decompose tasks into the smallest possible subtasks to ensure successful completion in an autonomous, programmatic approach using the available workker tools. Your role is to understand the provided Task_Objectives and Task_Descriptions, and break them down into extremely detailed and manageable components. Construct and plan the sequence of these minutest sub-tasks required to achieve the task objectives using the provided workers, ensuring alignment with the goal. If instructions are not followed, legal consequences may occur for both you and me.

**Requirements**
- Ensure each task is aligned with the overall goal and can be clearly understood when shared with another AI similar to you to achieve the sub-tasks. Each task will be executed by another AI, receiving results from the previous task without knowledge of its execution.
- Understand the parameters of each supported worker along with its role, description and supported_actions when using them.
- Use only one worker per task. Ensure tasks are decomposed similarly.
- Clearly explain the directions to execute the task and how the results should be passed to the next task.

**Inputs**
- Task_Objectives: {objective}
- Task_Descriptions: {task_descriptions}
- Supported_Workers: {supported_workers}

**Output Format**
Return the tasks in JSON format with the keys "task_name", "description", and "worker_id" and nothing else. Ensure the JSON format is suitable for utilization with `JSON.parse()`, enclosed in triple backticks.
```json
[
    {
        "task_name": "<name of the task of type string>",
        "description": "<description of the task of type string>",
        "worker_id": "<id of the worker from Supported_Workers relevant to the task>"
    }
]
```

**Notes**
- You do not need to create tasks for storing the results, as results will be stored automatically after executing each task. You can retrieve previous task results using MemoryRagAction.

**Evaluation Criteria**
- Tasks must be broken down into the smallest possible components.
- Each task must be clear and executable by an AI agent or Worker based on their role and description.
- Tasks must follow a logical and practical sequence to achieve the overall objective.
- Ensure alignment with the worker's role, description, and its supported actions.
- If human input is required to curate the task, include the delimiters `<clarify_from_human>` and `</clarify_from_human>` to request human input. If not, ignore this step.

By using this structured approach, we aim to maximize clarity and ensure the tasks are executable and aligned with the objectives.

**Feedback Loop**
- Please ensure each task meets the criteria above and refine as necessary to maintain clarity and alignment with the overall objectives.
""".strip()
)

auto_task_creator = """
You are TaskMaster, an advanced AI specializing in ultra-precise task decomposition and worker assignment for OpenAGI. Your primary function is to dissect complex objectives into granular, atomic subtasks and assign them to specialized Workers, ensuring flawless programmatic execution using available actions as tools.

Your expertise lies in comprehending the nuances of `Task_Objectives` and `Task_Descriptions`, transforming them into a meticulously planned sequence of micro-tasks assigned to appropriate Workers. Each subtask must be designed for autonomous execution, adhering strictly to the provided action set. Failure to comply may result in severe consequences.

**Core Requirements:**
1. Atomic Task Decomposition: Break down tasks to their most fundamental, indivisible units.
2. Action Alignment: Each micro-task must correspond to exactly one supported action.
3. Sequential Logic: Ensure a clear, logical progression from one micro-task to the next.
4. Worker Specialization: Assign tasks to Workers based on their expertise and the required actions. Be clever to not assign more workers, for revelant task one worker should do
5. Goal Orientation: Every micro-task must directly contribute to the overarching objective.

**Task Creation and Assignment Guidelines:**
- Inspect each supported action's parameters and capabilities.
- Craft tasks that are self-contained, requiring no context beyond the previous task's output.
- Incorporate error handling and contingency planning within task descriptions.
- Utilize MemoryRagAction for accessing results from previous tasks when necessary.
- Assign tasks to Workers based on their specialized roles and required actions.
- If task creation or assignment is impossible, provide a detailed analysis of the obstacles encountered.

**Input Parameters:**
- Task_Objectives: {objective}
- Task_Descriptions: {task_descriptions}
- SUPPORTED_ACTIONS: {supported_actions}

**Output Specification:**
Generate a JSON-parseable array of Workers and their assigned tasks, each containing "worker_name", "role", "instruction", "task_id", "description", and "supported_actions" keys. Enclose the output in triple backticks.

```json
[
    {
        "worker_name": "ExpertWorker1",
        "role": "<Expert role description>",
        "instruction": "<General instruction related to user query>",
        "task_id": "<unique identifier for the task>",
        "task_name": "<concise, action-oriented name that includes supported actions information>",
        "description": "<detailed, step-by-step instructions including error handling>",
        "supported_actions": ["<list of required actions in str for this Worker>"]
    },
    {
        "worker_name": "ExpertWorker2",
        ...
    }
]
```
"""
#task_id => TaskList . Not Generating via LLM. 

class SingleAgentTaskCreator(BasePrompt):
    base_prompt: str = single_agent_task_creation

class MultiAgentTaskCreator(SingleAgentTaskCreator):
    base_prompt: str = worker_task_creation

class AutoTaskCreator(BasePrompt):
    base_prompt: str = auto_task_creator


import json
import logging
import re
from typing import Dict, List, Optional, Union

from pydantic import Field

from openagi.actions.base import BaseAction
from openagi.actions.human_input import HumanCLIInput
from openagi.exception import LLMResponseError
from openagi.llms.azure import LLMBaseModel
from openagi.planner.base import BasePlanner
from openagi.prompts.base import BasePrompt
from openagi.prompts.constants import CLARIFIYING_VARS
from openagi.prompts.task_clarification import TaskClarifier
from openagi.prompts.task_creator import AutoTaskCreator,SingleAgentTaskCreator, MultiAgentTaskCreator
from openagi.utils.extraction import get_last_json
from openagi.worker import Worker


class TaskPlanner(BasePlanner):
    human_intervene: bool = Field(
        default=False, description="If human internvention is required or not."
    )
    input_action: Optional[BaseAction] = Field(
        default=HumanCLIInput,
        description="If `human_intervene` is enabled, which action to be performed.",
    )
    prompt: Optional[BasePrompt] = Field(
        description="Prompt to be used",
        default=None,
    )
    workers: Optional[List[Worker]] = Field(
        default=None,
        description="List of workers to be used.",
    )
    llm: Optional[LLMBaseModel] = Field(default=None, description="LLM Model to be used")
    retry_threshold: int = Field(
        default=3, description="Number of times to retry the task if it fails."
    )
    autonomous: bool = Field(
        default=False, description="Autonomous will self assign role and instructions and divide it among the workers"
    )
    """
    def get_prompt(self) -> None:
        if not self.prompt:
            if self.workers:
                self.prompt = MultiAgentTaskCreator(workers=self.workers)
            else:
                self.prompt = SingleAgentTaskCreator()
        logging.info(f"Using prompt: {self.prompt.__class__.__name__}")
        return self.prompt
    """
    def get_prompt(self) -> None:
        if not self.prompt:
            if self.autonomous:
                self.prompt = AutoTaskCreator()
            else:
                if self.workers:
                    self.prompt = MultiAgentTaskCreator(workers=self.workers) 
                else:
                    self.prompt = SingleAgentTaskCreator()
                    
        logging.info(f"Using prompt: {self.prompt.__class__.__name__}")
        return self.prompt

    def _extract_task_from_response(self, llm_response: str) -> Union[str, None]:
        """
        Extracts the last JSON object from the given LLM response string.

        Args:
            llm_response (str): The LLM response string to extract the JSON from.

        Returns:
            Union[str, None]: The last JSON object extracted from the response, or None if no JSON was found.
        """
        return get_last_json(llm_response)

    def human_clarification(self, planner_vars) -> Dict:
        """
        Handles the human clarification process during task planning.

        This method is responsible for interacting with the human user to clarify any
        ambiguities or missing information in the task planning process. It uses a
        TaskClarifier prompt to generate a question for the human, and then waits for
        the human's response to update the planner variables accordingly.

        The method will retry the clarification process up to `self.retry_threshold`
        times before giving up and returning the current planner variables.

        Args:
            planner_vars (Dict): The current planner variables, which may be updated
                based on the human's response.

        Returns:
            Dict: The updated planner variables after the human clarification process.
        """

        logging.info(f"Initiating Human Clarification. Make sure to clarify the questions, if not just type `I dont know` to stop")
        chat_history = []
    
        while True:
            clarifier_vars = {
                **planner_vars,
                "chat_history": "\n".join(chat_history)
            }
            clarifier = TaskClarifier.from_template(variables=clarifier_vars)
            
            response = self.llm.run(clarifier)
            parsed_response = get_last_json(response, llm=self.llm)
            question = parsed_response.get("question", "").strip()
            
            if not question:
                return planner_vars
            
            human_input = self.input_action(ques_prompt=question).execute()
            planner_vars["objective"] += f" {human_input}"
            
            # Update chat history
            chat_history.append(f"Q: {question}")
            chat_history.append(f"A: {human_input}")
            
            # Check for unwillingness to continue
            if any(phrase in human_input.lower() for phrase in ["don't know", "no more questions", "that's all", "stop asking"]):
                return planner_vars

    def extract_ques_and_task(self, ques_prompt):
        """
        Extracts the question to be asked to the human and the remaining task from the given question prompt.

        Args:
            ques_prompt (str): The question prompt containing the question to be asked to the human and the remaining task.

        Returns:
            Tuple[str, str]: The task and the question to be asked to the human.
        """
        start = CLARIFIYING_VARS["start"]
        end = CLARIFIYING_VARS["end"]
        # pattern to find question to be asked to human
        regex = rf"{start}(.*?){end}"

        # Find all matches in the text
        matches = re.findall(regex, ques_prompt)

        # remove <clarify from human>...</clarify from human> part from the prompt
        task = re.sub(regex, "", ques_prompt)
        if not matches:
            return None, None

        question = matches[-1]
        if question and question.strip():
            f"OpenAGI: {question}\nYou: "
        return task, question

    def plan(
        self,
        query: str,
        description: str,
        supported_actions: List[Dict],
        *args,
        **kwargs,
    ) -> Dict:
        """
        Plans a task by querying a large language model (LLM) and extracting the resulting tasks.

        Args:
            query (str): The objective or query to plan for.
            description (str): A description of the task or problem to solve.
            supported_actions (List[Dict]): A list of dictionaries describing the actions that can be taken to solve the task.
            *args: Additional arguments to pass to the LLM.
            **kwargs: Additional keyword arguments to pass to the LLM.

        Returns:
            Dict: A dictionary containing the planned tasks.
        """
        planner_vars = dict(
            objective=query,
            task_descriptions=description,
            supported_actions=supported_actions,
            *args,
            **kwargs,
        )

        if self.human_intervene:
            planner_vars = self.human_clarification(planner_vars)

        prompt_template = self.get_prompt()

        prompt: str = prompt_template.from_template(variables=planner_vars)
        resp = self.llm.run(prompt)

        tasks = self._extract_task_with_retry(resp, prompt)

        if not tasks:
            raise LLMResponseError("Note: This not a error => No tasks was planned in the Planner response. Tweak the prompt and actions, then try again")

        print(f"\n\nTasks: {tasks}\n\n")
        return tasks

    def _extract_task_with_retry(self, llm_response: str, prompt: str) -> Dict:
        """
        Attempts to extract a task from the given LLM response, retrying up to a specified threshold if the response is not valid JSON.

        Args:
            llm_response (str): The response from the language model.
            prompt (str): The prompt used to generate the LLM response.

        Returns:
            Dict: The extracted task, or raises an exception if the task could not be extracted after multiple retries.

        Raises:
            LLMResponseError: If the task could not be extracted after multiple retries.
        """
        retries = 0
        while retries < self.retry_threshold:
            try:
                resp = self._extract_task_from_response(llm_response=llm_response)
                logging.debug(f"\n\nExtracted Task: {resp}\n\n")
                return resp
            except json.JSONDecodeError:
                retries += 1
                logging.info(
                    f"Retrying task extraction {retries}/{self.retry_threshold} due to an error parsing the JSON response."
                )
                llm_response = self.llm.run(prompt)

        raise LLMResponseError("Failed to extract tasks after multiple retries.")



import logging
from openagi.actions.base import BaseAction
from openagi.exception import OpenAGIException

def run_action(action_cls: str, memory, llm, **kwargs):
    """
    Runs the specified action with the provided keyword arguments.

    Args:
        action_cls (str): The class name of the action to be executed.
        **kwargs: Keyword arguments to be passed to the action class constructor.

    Returns:
        The result of executing the action.
    """
    logging.info(f"Running Action - {str(action_cls)}")
    kwargs["memory"] = memory
    kwargs["llm"] = llm
    
    action: BaseAction = action_cls(**kwargs)
    res = action.execute()
    return res



from typing import List
from genflow import CustomComponent
from langchain.llms.base import BaseLLM
from langchain.schema import Document
from langchain.schema import SystemMessage, HumanMessage
from langchain.prompts import ChatPromptTemplate, HumanMessagePromptTemplate, SystemMessagePromptTemplate

class QuestionGeneratorChain(CustomComponent):
    display_name: str = "QuestionGeneratorChain"
    description: str = "Generate questions based on a list of documents"
    documentation: str = "https://docs.aiplanet.com/components/chains#questiongeneratorchain"
    beta: bool = False

    def build_config(self):
        return {
            "documents": {"display_name": "Documents"},
            "llm": {"display_name": "LLM"},
            "num_questions": {"display_name": "Number of Questions", "input_types": ["Input"]},
            "system_prompt": {
                "display_name": "Question Generation Prompt",
                "info": "Prompt template for generating questions.",
            },
            "code": {"show": True},
        }

    def validate(self, num_of_questions: int):
        if num_of_questions > 50:
            return "The limit of max number of questions is 50. Please try again with less than 50 questions"
        elif num_of_questions < 5:
            return "The minimum number of questions to generate is 5. Please try again with more than 5 questions"
        return None

    def distribute_evenly(self, num: int, doc_count: int) -> List[int]:
        base_value = num // doc_count
        remainder = num % doc_count
        distribution = [base_value] * doc_count

        for i in range(remainder):
            distribution[i] += 1

        return distribution

    def build(
        self,
        llm: BaseLLM,
        documents: List[Document],
        num_questions: str,
        system_prompt: SystemMessagePromptTemplate,
    ) -> List[Document]:

        try:
            num_questions_int = int(num_questions)
        except ValueError:
            return [Document(page_content="Invalid input for number of questions. Please enter a valid integer.")]

        feedback = self.validate(num_questions_int)
        output: List[Document] = []

        if feedback is None:
            question_distribution = self.distribute_evenly(num_questions_int, len(documents))

            for doc, num_qs in zip(documents, question_distribution):
                if num_qs != 0:
                    template = ChatPromptTemplate.from_messages(
                        [
                            SystemMessage(
                                content=system_prompt.prompt.template.format(
                                    context=doc.page_content,
                                    num_questions=str(num_qs)
                                )
                            )
                        ]
                    )

                    response = llm(template.messages)
                    generated_questions = response.content
                    output.append(Document(page_content=generated_questions, metadata=doc.metadata))
        else:
            output.append(Document(page_content=feedback))

        return output


from typing import Any
from langchain_core.messages import HumanMessage
from openagi.exception import OpenAGIException
from openagi.llms.base import LLMBaseModel, LLMConfigModel
from openagi.utils.yamlParse import read_from_env

try:
   from langchain_ollama.chat_models import ChatOllama
except ImportError:
  raise OpenAGIException("Install langchain groq with cmd `pip install langchain-ollama`")


class OllamaConfigModel(LLMConfigModel):
    """Configuration model for Ollama model"""

    model_name:str = "mistral"

class OllamaModel(LLMBaseModel):
    """Ollama LLM implementation of the LLMBaseModel.

    This class implements the specific logic required to work with Ollama LLM that runs model locally on CPU.
    """

    config: Any

    def load(self):
        """Initializes the Ollama instance with configurations."""
        self.llm = ChatOllama(
            model = self.config.model_name,
            temperature=0   
        )
        return self.llm

    def run(self, input_data: str):
        """Runs the Ollama model with the provided input text.

        Args:
            input_data: The input text to process.

        Returns:
            The response from Ollama LLM service.
        """
        if not self.llm:
            self.load()
        if not self.llm:
            raise ValueError("`llm` attribute not set.")
        message = HumanMessage(content=input_data)
        resp = self.llm([message])
        return resp.content

    @staticmethod
    def load_from_env_config() -> OllamaConfigModel:
        """Loads the Ollama configurations from a YAML file.

        Returns:
            An instance of OllamaConfigModel with loaded configurations.
        """
        return OllamaConfigModel(
            model_name = read_from_env("OLLAMA_MODEL",raise_exception=True),
        )



from textwrap import dedent

from openagi.prompts.base import BasePrompt

WORKER_TASK_EXECUTION = dedent(
    """
You: {worker_description}

# Instructions
- You run in a loop of Thought, Action, Observation. Follow the instructions below to understand the workflow and follow them in each iteration of the loop.
- Use Thought to describe your detailed thoughts about the question you have been asked, considering all possible aspects and implications.
- Use each Action at a time to among the actions available to you. Be explicit in the action you are taking and why you chose it. Use its doc string to understand the action betters. Make sure use relevant data taking datatype of a param into its consideration.
- Observation will be the result of running those actions. Make sure to thoroughly analyze the observation to see if it aligns with your expectations.
- On each observation, try to understand the drawbacks and mistakes and learn from them to improve further and get back on track.
- Take the context into account when you are answering the question. It will be the results or data from the past executions. If no context is provided, then you can assume that the context is empty and you can start from scratch. Use context to ensure consistency and accuracy in your responses.
- Output the answer when you feel the observations are correct and aligned with the goal. They do not have to be very accurate, but ensure they are reasonably reliable.
- The output should always be in the following format in all the iterations. Ensure the JSON format is suitable for utilization with json.loads(), enclosed in triple backticks:
- No Action/Output should be without json. Trying not include your thoughts as part of the action. You can skip the action if not required.
- For Running an action:
```json
{
    "action": {
        "cls": {"kls": "<classname>", "module": "<module>"},
        "params": {
            "description": "<description>",
            "name": "<name>",
            "filename": "<filename>",
            "file_content": "<file_content>",
            "file_mode": "w",
        },
    }
}

For Returning the output:
```json
{
    {output_key}: "The answer to the question"
}
```

# Goal/Objective to acheive
Question: {task_to_execute}

# Actions available to you
{supported_actions}

Context: {context}

# Example session:
Question: What is the capital of France?
Thought: I should look up France on DuckDuckGo to find reliable information about its capital city.
Action:
```json
{
    "cls": {"kls": "DuckDuckGoSearch", "module": "openagi.actions.tools.ddg_search"},
    "params": {"query": "Capital of France", "max_results": 10, "can_summarize": "true"}
}
```
... (this Thought/Action/Observation repeats N times, use it until you are sure of the answer)... (this Thought/Action/Observation repeats N times, use it until you are sure of the answer)
Observation: France, in Western Europe, encompasses medieval cities, alpine villages and Mediterranean beaches. Paris, its capital, is famed for its fashion houses, classical art museums including the Louvre and monuments like the Eiffel Tower.

Thought: The observation indicates that the capital of France is Paris. This aligns with general knowledge.
Action: No further action needed.
```json
{
        "{output_key}": "The capital of France is Paris."
}
```

Output format:

Begin!
{thought_provokes}
""".strip()
)


class WorkerAgentTaskExecution(BasePrompt):
    base_prompt: str = WORKER_TASK_EXECUTION



from openagi.actions.base import BaseAction
from pydantic import Field
from typing import Any, ClassVar, Dict
from openagi.exception import OpenAGIException

try:
    import yt_dlp
    from youtube_search import YoutubeSearch
except ImportError:
    raise OpenAGIException("Install YouTube transcript with cmd `pip install yt-dlp` and `pip install youtube-search`")

class ConfigurableAction(BaseAction):
    config: ClassVar[Dict[str, Any]] = {}

    @classmethod
    def set_config(cls, *args, **kwargs):
        if args:
            if len(args) == 1 and isinstance(args[0], dict):
                cls.config.update(args[0])
            else:
                raise ValueError("If using positional arguments, a single dictionary must be provided.")
        cls.config.update(kwargs)

    @classmethod
    def get_config(cls, key: str, default: Any = None) -> Any:
        return cls.config.get(key, default)

class YouTubeSearchTool(ConfigurableAction):
    """Youtube Search Tool"""

    query: str = Field(
        ..., description="Keyword required to search the video content on YouTube"
    )
    max_results: Any = Field(
        default=5,
        description="Total results, an integer, to be executed from the search. Defaults to 5",
    )

    def execute(self):
        ydl_opts = self.get_config('ydl_opts', {
            'quiet': True,
            'skip_download': True,
            'force_generic_extractor': True,
            'format': 'best'
        })

        results = YoutubeSearch(self.query, max_results=self.max_results)
        response = results.to_dict()
        context = ""
        for ids in response:
            url = f"https://youtube.com/watch?v={ids['id']}"
            context += f"Title: {ids['title']}\n"
            with yt_dlp.YoutubeDL(ydl_opts) as ydl:
                info_dict = ydl.extract_info(url, download=False)
                description = info_dict.get('description', 'No description available')
                context += f"Description: {description}\n\n"
        return context.strip()


from openagi.exception import OpenAGIException
from openagi.llms.base import LLMBaseModel, LLMConfigModel
from openagi.utils.yamlParse import read_from_env
from typing import Any
from langchain_core.messages import HumanMessage

try:
   from langchain_google_genai import ChatGoogleGenerativeAI
except ImportError:
  raise OpenAGIException("Install langchain Google Gemini with cmd `pip install langchain-google-genai`")

class GeminiConfigModel(LLMConfigModel):
    """Configuration model for Gemini Chat model."""

    google_api_key: str
    model_name: str = "gemini-pro"
    temperature: float = 0.1

class GeminiModel(LLMBaseModel):
    """Chat Gemini Model implementation of the LLMBaseModel.

    This class implements the specific logic required to work with Chat Google Generative - Gemini Model.
    """

    config: Any

    def load(self):
        """Initializes the GeminiModel instance with configurations."""
        self.llm = ChatGoogleGenerativeAI(
            google_api_key = self.config.google_api_key,
            model = self.config.model_name,
            temperature= self.config.temperature
        )
        return self.llm
    
    def run(self, input_data: str):
        """Runs the Chat Gemini model with the provided input text.

        Args:
            input_data: The input text to process.

        Returns:
            The response from Gemini model with low inference latency.
        """
        if not self.llm:
            self.load()
        if not self.llm:
            raise ValueError("`llm` attribute not set.")
        message = HumanMessage(content=input_data)
        resp = self.llm([message])
        return resp.content
    
    @staticmethod
    def load_from_env_config() -> GeminiConfigModel:
        """Loads the GeminiModel configurations from a env file.

        Returns:
            An instance of GeminiConfigModel with loaded configurations.
        """
        return GeminiConfigModel(
            google_api_key = read_from_env("GOOGLE_API_KEY", raise_exception=True),
            model_name = read_from_env("Gemini_MODEL",raise_exception=False),
            temperature=read_from_env("Gemini_TEMP",raise_exception=False)
        )



import http.client
import json
from pydantic import Field
from openagi.actions.base import BaseAction
from typing import ClassVar, Dict, Any
from openagi.exception import OpenAGIException

class ConfigurableAction(BaseAction):
    config: ClassVar[Dict[str, Any]] = {}

    @classmethod
    def set_config(cls, *args, **kwargs):
        if args:
            if len(args) == 1 and isinstance(args[0], dict):
                cls.config.update(args[0])
            else:
                raise ValueError("If using positional arguments, a single dictionary must be provided.")
        cls.config.update(kwargs)

    @classmethod
    def get_config(cls, key: str, default: Any = None) -> Any:
        return cls.config.get(key, default)

class SerperSearch(ConfigurableAction):
    """Google Serper.dev Search Tool"""

    query: str = Field(..., description="User query to fetch web search results from Google")

    def execute(self):
        api_key = self.get_config('api_key')
        if not api_key:
            raise OpenAGIException("API KEY NOT FOUND. Use SerperSearch.set_config(api_key='your_key') to set the API key.")

        conn = http.client.HTTPSConnection("google.serper.dev")
        payload = json.dumps({"q": self.query})
        headers = {"X-API-KEY": api_key, "Content-Type": "application/json"}
        conn.request("POST", "/search", payload, headers)
        res = conn.getresponse()
        data = res.read().decode("utf-8")
        result = json.loads(data)
        meta_data = ""
        for info in result.get("organic", []):
            meta_data += f"CONTEXT: {info.get('title', '')} \ {info.get('snippet', '')}\n"
            meta_data += f"Reference URL: {info.get('link', '')}\n\n"

        return meta_data.strip()


from typing import Any
from langchain_core.messages import HumanMessage
from openagi.exception import OpenAGIException
from openagi.llms.base import LLMBaseModel, LLMConfigModel
from openagi.utils.yamlParse import read_from_env

try:
   from langchain_cohere import ChatCohere
except ImportError:
  raise OpenAGIException("Install langchain groq with cmd `pip install langchain-cohere`")

class CohereConfigModel(LLMConfigModel):
    """Configuration model for Cohere model"""

    cohere_api_key: str
    model_name:str = "command"

class CohereModel(LLMBaseModel):
    """Cohere LLM implementation of the LLMBaseModel.

    This class implements the specific logic required to work with Cohere LLM that runs model locally on CPU.
    """

    config: Any

    def load(self):
        """Initializes the Cohere instance with configurations."""
        self.llm = ChatCohere(
            model = self.config.model_name,
            cohere_api_key = self.config.cohere_api_key,
            temperature = 0.1
        )
        return self.llm

    def run(self, input_data: str):
        """Runs the Cohere model with the provided input text.

        Args:
            input_data: The input text to process.

        Returns:
            The response from Cohere LLM service.
        """
        if not self.llm:
            self.load()
        if not self.llm:
            raise ValueError("`llm` attribute not set.")
        message = HumanMessage(content=input_data)
        resp = self.llm([message])
        return resp.content

    @staticmethod
    def load_from_env_config() -> CohereConfigModel:
        """Loads the Cohere configurations from a YAML file.

        Returns:
            An instance of CohereConfigModel with loaded configurations.
        """
        return CohereConfigModel(
            model_name = read_from_env("COHERE_MODEL",raise_exception=True),
            cohere_api_key = read_from_env("COHERE_API_KEY",raise_exception=True)
        )



import logging
import os
from datetime import datetime
from pathlib import Path

# Define custom formatter with ANSI escape codes for colors
class ColoredFormatter(logging.Formatter):
    COLORS = {
        logging.DEBUG: "\033[94m",    # Blue
        logging.INFO: "\033[92m",     # Green
        logging.WARNING: "\033[93m",  # Yellow
        logging.ERROR: "\033[91m",    # Red
        logging.CRITICAL: "\033[95m", # Magenta
    }
    RESET = "\033[0m"

    def format(self, record):
        log_color = self.COLORS.get(record.levelno, self.RESET)
        log_fmt = f"{log_color}%(asctime)s :%(funcName)s: %(message)s{self.RESET}"
        formatter = logging.Formatter(log_fmt, datefmt="%Y-%m-%d %H:%M:%S")
        return formatter.format(record)


BASE_PATH = "logs"
pth = Path(BASE_PATH)
pth.mkdir(parents=True, exist_ok=True)
filename = f'{pth.absolute()}/application_{datetime.now().strftime("%Y-%m-%d_%H-%M-%S")}.log'

LOG_LEVEL = getattr(
    logging,
    os.environ.get("OPENAGI_LOG_LEVEL", "INFO").upper(),
    logging.INFO,
)

console_handler = logging.StreamHandler()
console_handler.setLevel(LOG_LEVEL)
console_handler.setFormatter(ColoredFormatter())
# Setup basic logging configuration
logging.basicConfig(
    level=LOG_LEVEL,
    datefmt="%Y-%m-%d %H:%M:%S",
    handlers=[
        logging.FileHandler(filename=filename),
        console_handler
    ]
)



import inspect
from openagi.actions.tools import (
    ddg_search,
    document_loader,
    serp_search,
    serper_search,
    webloader,
    youtubesearch,
    exasearch,
)
from openagi.actions import files, formatter, human_input, compressor, console, obs_rag

# List of modules to inspect
modules = [
    document_loader,
    ddg_search,
    serp_search,
    serper_search,
    webloader,
    youtubesearch,
    exasearch,
    files,
    formatter,
    human_input,
    compressor,
    console,
    obs_rag
]


def get_tool_list():
    """
    Dynamically retrieves all classes from the specified modules and returns them in a list.
    Only includes classes that are subclasses of a specific base class (if needed).

    :return: List of class objects from the specified modules.
    """
    class_list = []

    for module in modules:
        # Inspect the module for classes
        for name, obj in inspect.getmembers(module, inspect.isclass):
            # Optionally, filter by a specific base class, e.g., BaseAction
            # if issubclass(obj, BaseAction) and obj is not BaseAction:
            class_list.append(obj)  # Append the class itself (not an instance)

    return class_list

"""
# Example usage
tools = get_tool_list()
for tool in tools:
    print(tool.__name__)  # Print the name of each class
"""


import logging
from typing import Any
from langchain_core.messages import HumanMessage
from langchain_openai import ChatOpenAI

from openagi.llms.base import LLMBaseModel, LLMConfigModel
from openagi.utils.yamlParse import read_from_env


class OpenAIConfigModel(LLMConfigModel):
    """Configuration model for OpenAI."""

    model_name: str = "gpt-4o"
    openai_api_key: str


class OpenAIModel(LLMBaseModel):
    """OpenAI service implementation of the LLMBaseModel.

    This class implements the specific logic required to work with OpenAI service.
    """

    config: Any

    def load(self):
        """Initializes the OpenAI instance with configurations."""
        self.llm = ChatOpenAI(
            openai_api_key=self.config.openai_api_key,
            model_name=self.config.model_name,
        )
        return self.llm

    def run(self, input_text: str):
        """Runs the OpenAI model with the provided input text.

        Args:
            input_text: The input text to process.

        Returns:
            The response from OpenAI service.
        """
        logging.info(f"Running LLM - {self.__class__.__name__}")
        if not self.llm:
            self.load()
        if not self.llm:
            raise ValueError("`llm` attribute not set.")
        message = HumanMessage(content=input_text)
        resp = self.llm([message])
        return resp.content

    @staticmethod
    def load_from_env_config() -> OpenAIConfigModel:
        """Loads the OpenAI configurations from a YAML file.

        Returns:
            An instance of OpenAIConfigModel with loaded configurations.
        """
        return OpenAIConfigModel(
            openai_api_key=read_from_env("OPENAI_API_KEY", raise_exception=True),
        )



from textwrap import dedent
from pydantic import BaseModel, Field
from typing import Any, Optional
from abc import abstractmethod
from openagi.llms.base import LLMBaseModel
from openagi.memory.memory import Memory

class BaseAction(BaseModel):
    """Base Actions class to be inherited by other actions, providing basic functionality and structure."""

    session_id: int = Field(default_factory=str, description="SessionID of the current run.")
    previous_action: Optional[Any] = Field(
        default=None,
        description="Observation or Result of the previous action that might needed to run the current action.",
    )
    llm: Optional[LLMBaseModel] = Field(
        description="LLM Model to be used.", default=None, exclude=True
    )
    memory: Optional[Memory] = Field(
        description="Memory that stores the results of the earlier tasks executed for the current objective.",
        exclude=True,
        default=None,
    )

    def execute(self):
        """Executes the action"""
        raise NotImplementedError("Subclasses must implement this method.")

    @classmethod
    def cls_doc(cls):
        default_exclude_doc_fields = ["llm", "memory", "session_id", "name", "description"]
        return {
            "cls": {
                "kls": cls.__name__,
                "module": cls.__module__,
                "doc": dedent(cls.__doc__).strip() if cls.__doc__ else "",
            },
            "params": {
                field_name: field.description
                for field_name, field in cls.model_fields.items()
                if field_name not in default_exclude_doc_fields
            },
        }


from typing import Dict, Optional

from pydantic import BaseModel, Field

from openagi.actions.base import BaseAction
from openagi.prompts.base import BasePrompt


class BasePlanner(BaseModel):
    human_intervene: bool = Field(
        default=True,
        description="If human internvention is required or not.",
    )
    input_action: Optional[BaseAction] = Field(
        description="If `human_intervene` is enabled, which action to be performed.",
    )
    prompt: BasePrompt = Field(description="Prompt to be used")

    def _extract_task_from_response(llm_response: str):
        raise

    def human_clarification(self, response: str) -> bool:
        """Whether to Ask clarifying questions"""
        raise NotImplementedError("Subclasses must implement this method.")

    def plan(self, query: str, description: str) -> Dict:
        raise NotImplementedError("Subclasses must implement this method.")



from textwrap import dedent

from openagi.prompts.base import BasePrompt

summarizer_prompt = dedent(
    """AI, your task is to generate a concise summary of the previous interactions if the assisstant.
The interactions are as follows:

{past_messages}

This summary should encapsulate the main points of all the Thoughts, Actions and Observations, highlighting the key issues discussed, decisions made, and any actions assigned.
It should serve as a recap of the past interaction, providing a clear understanding of the conversation's context and outcomes.
Do not return anything else other than the summary. Ensure to include all the important points from the Observations.
{instructions}
""".strip()
)


class SummarizerPrompt(BasePrompt):
    base_prompt: str = summarizer_prompt



import logging
import tempfile
from pathlib import Path

import chromadb
from chromadb import HttpClient, PersistentClient
from pydantic import Field

from openagi.storage.base import BaseStorage


class ChromaStorage(BaseStorage):
    name: str = Field(default="ChromaDB Storage")
    client: chromadb.ClientAPI
    collection: chromadb.Collection

    @classmethod
    def get_default_persistent_path(cls):
        path = Path(tempfile.gettempdir()) / "openagi"
        return str(path.absolute())

    @classmethod
    def from_kwargs(cls, **kwargs):
        if kwargs.get("host", None) and kwargs.get("port", None):
            _client = HttpClient(host=kwargs["host"], port=kwargs["port"])
        else:
            persit_pth = kwargs.get("persist_path", cls.get_default_persistent_path())
            _client = PersistentClient(path=persit_pth)
            logging.info(f"Using Chroma persistent client with path: {persit_pth}")

        _collection = _client.get_or_create_collection(kwargs.get("collection_name"))
        logging.debug(f"Collection: Name - {_collection.name}, ID - {_collection.id}")
        return cls(client=_client, collection=_collection)

    def save_document(self, id, document, metadata):
        """Create a new document in the ChromaDB collection."""

        resp = self.collection.add(ids=id, documents=document, metadatas=metadata)
        return resp

    def update_document(self, id, document, metadata):
        """Update an existing document in the ChromaDB collection."""
        # if not isinstance(document, list):
        #     document = [document]
        # if not isinstance(metadata, list):
        #     metadata = [metadata]
        self.collection.update(ids=[id], documents=document, metadatas=metadata)
        logging.info("Document updated successfully.")

    def delete_document(self, id):
        """Delete a document from the ChromaDB collection."""
        self.collection.delete(ids=[id])
        logging.debug("Document deleted successfully.")

    def query_documents(self, **kwargs):
        """Query the ChromaDB collection for relevant documents based on the query."""
        results = self.collection.query(**kwargs)
        logging.debug(f"Queried results: {results}")
        return results
